#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:class:`QComboBox`-based simulation configuration widget subclasses.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, Signal  #, Slot
from PySide2.QtWidgets import QComboBox
# from betse.util.io.log import logs
from betse.util.type.iterable import iterables
from betse.util.type.mapping import mappings
from betse.util.type.types import type_check, EnumClassType
from betsee.guiexception import BetseePySideComboBoxException
from betsee.gui.simconf.stack.widget.abc.guisimconfwdgeditenum import (
    QBetseeSimConfEditEnumWidgetMixin)
from collections import OrderedDict

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfEnumComboBox(
    QBetseeSimConfEditEnumWidgetMixin, QComboBox):
    '''
    Simulation configuration-specific combo box widget, permitting high-level
    enumeration members backed by low-level raw strings in external simulation
    configuration files to be interactively edited.

    To guarantee a one-to-one correspondence between the currently selected
    combo box item and the enumeration constraining these items, this combo box
    is non-editable (i.e., item text is *not* manually editable).

    See Also
    ----------
    :class:`QBetseeSimConfEnumRadioButtonGroup`
        Alternative simulation configuration-specific widget subclass similarly
        permitting high-level enumeration members to be interactively edited.
        That subclass is both less cumbersome to initialize *and* preferable from
        the user experience (UX) perspective for sufficiently small enumerations
        (e.g., containing five or fewer members).
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed arguments.
        super().__init__(*args, **kwargs)

        # Disable editability, preventing users from manually editing item text.
        # (Though editability is technically disabled by our superclass by
        # default, this is sufficiently important to warrant disabling anyway.)
        self.setEditable(False)


    @type_check
    def init(
        self, enum_member_to_item_text: OrderedDict, *args, **kwargs) -> None:
        '''
        Finalize the initialization of this widget.

        Parameters
        ----------
        enum_member_to_item_text : OrderedDict
            Ordered dictionary mapping from each member of the enumeration
            encapsulated by the passed ``sim_conf_alias`` parameter to the
            human-readable text of the combo box item describing that member.
            The dictionary ordering of these enumeration members exactly defines
            the order in which the corresponding combo box items are listed.

        All remaining parameters are passed as is to the superclass method.

        Raises
        ----------
        BetseMappingException
            If this dictionary is *not* safely invertible (i.e., if any value of
            this dictionary is non-uniquely assigned to two or more keys).
        BetseePySideComboBoxException
            If the number of members in this enumeration differs from the number
            of members mapped by (i.e., of keys in) this dictionary.
        '''

        # Dictionary mapping from each enumeration member to the corresponding
        # mutually exclusive value displayed by this widget -- which, in this
        # case, is the 0-based index of that combo box item.
        enum_member_to_widget_value = {
            enum_member: item_index
            for item_index, enum_member in enumerate(
                enum_member_to_item_text.keys())
        }

        # Initialize our superclass with all such parameters.
        super().init(
            *args,
            enum_member_to_widget_value=enum_member_to_widget_value,
            **kwargs)

        # Sequence of the human-readable text of all combo box items.
        #
        # Since a view of ordered dictionary values is a valid sequence, the
        # QComboBox.addItems() method passed this sequence below should
        # technically accept this view *WITHOUT* requiring explicit coercion
        # into a tuple. It does not, raising the following exception on
        # attempting to do so:
        #
        #    TypeError: 'PySide2.QtWidgets.QComboBox.addItems' called with wrong argument types:
        #      PySide2.QtWidgets.QComboBox.addItems(ValuesView)
        #    Supported signatures:
        #      PySide2.QtWidgets.QComboBox.addItems(QStringList)
        #
        # Presumably, the underlying bindings generator (i.e., Shiboken2) only
        # implicitly coerces tuples and lists into "QStringList" instances.
        # Curiously, generic sequences appear to remain unsupported.
        items_text = tuple(enum_member_to_item_text.values())

        # Populate the contents of this combo box from this sequence.
        self.addItems(items_text)

    # ..................{ MIXIN ~ property : read-only       }..................
    @property
    def undo_synopsis(self) -> str:

        return QCoreApplication.translate(
            'QBetseeSimConfEnumComboBox', 'changes to a combo box')


    @property
    def _finalize_widget_change_signal(self) -> Signal:

        return self.currentIndexChanged

    # ..................{ MIXIN ~ property : value           }..................
    @property
    def widget_value(self) -> int:

        return self.currentIndex()


    @widget_value.setter
    @type_check
    def widget_value(self, widget_value: int) -> None:

        # Set this widget's displayed value to the passed value by calling the
        # setCurrentIndex() method of our superclass rather than this subclass,
        # preventing infinite recursion. (See the superclass method docstring.)
        super().setCurrentIndex(widget_value)


    def _reset_widget_value(self) -> None:

        # Reset this combo box by arbitrarily selecting its first item. Note
        # this is distinct from the clear() method defined by our superclass,
        # removing all existing items from this combo box. Calling that method
        # would be highly undesirable here.
        self.widget_value = 0



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **worker thread pool** (i.e., platform-portable, pure-Qt,
:class:`QThreadPool`-based container of one or more threads, each working
exactly one :class:`QRunnable`-based worker at a given time) classes.
'''

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QThreadPool
from betse.util.io.log import logs
from betse.util.type.iterable import iterables
from betse.util.type.types import type_check, IterableTypes
from betsee.guiexception import BetseePySideThreadException
from betsee.util.type.guitype import QThreadPoolOrNoneTypes
from betsee.util.thread.pool.guipoolwork import QBetseeThreadPoolWorker

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_if_working(thread_pool: QThreadPoolOrNoneTypes = None) -> None:
    '''
    Raise an exception if one or more workers are currently working in one or
    more non-idle threads of the passed thread pool.

    Parameters
    ----------
    thread_pool : QThreadPoolOrNoneTypes
        Thread pool to run this worker in. Defaults to ``None``, in which case
        the singleton thread pool returned by the :func:`get_thread_pool`
        function is defaulted to.

    Raises
    ----------
    BetseePySideThreadException
        If one or more workers are currently working in one or more non-idle
        threads of this thread pool.

    See Also
    ----------
    :func:`is_working`
        Further details.
    '''

    # If this thread pool is currently working, raise an exception.
    if is_working(thread_pool):
        # Number of workers currently working in this thread pool.
        worker_count = get_worker_count(thread_pool)

        # Raise an exception synopsizing the state of this thread pool.
        raise BetseePySideThreadException(
            'Thread pool contains {} working workers.'.format(worker_count))

# ....................{ TESTERS                           }....................
@type_check
def is_working(thread_pool: QThreadPoolOrNoneTypes = None) -> bool:
    '''
    ``True`` only if one or more workers are currently working in one or more
    non-idle threads of the passed thread pool.

    Parameters
    ----------
    thread_pool : QThreadPoolOrNoneTypes
        Thread pool to run this worker in. Defaults to ``None``, in which case
        the singleton thread pool returned by the :func:`get_thread_pool`
        function is defaulted to.
    '''

    # Return true only if this thread pool has at least one non-idle thread.
    return get_worker_count(thread_pool) > 0

# ....................{ GETTERS                           }....................
def get_thread_pool() -> QThreadPool:
    '''
    Singleton **worker thread pool** (i.e., platform-portable, pure-Qt,
    :class:`QThreadPool`-based container of one or more threads, each working
    exactly one :class:`QRunnable`-based worker at a given time).

    This singleton is globally reusable across the entire application.
    '''

    return QThreadPool.globalInstance()


@type_check
def get_worker_count(thread_pool: QThreadPoolOrNoneTypes = None) -> int:
    '''
    Number of workers currently working in non-idle threads of the passed
    thread pool.

    Parameters
    ----------
    thread_pool : QThreadPoolOrNoneTypes
        Thread pool to inspect the wokers of. Defaults to ``None``, in which
        case the singleton thread pool returned by the :func:`get_thread_pool`
        function is defaulted to.
    '''

    # Default this thread pool to the singleton thread pool if needed.
    if thread_pool is None:
        thread_pool = get_thread_pool()

    # Return the number of non-idle threads in this thread pool.
    return thread_pool.activeThreadCount()

# ....................{ RUNNERS                           }....................
@type_check
def start_worker(
    worker: QBetseeThreadPoolWorker,
    thread_pool: QThreadPoolOrNoneTypes = None,
) -> None:
    '''
    Start the passed thread pool worker in the passed thread pool.

    Specifically, this function:

    * If this thread pool contains at least one idle thread:
      * Moves this worker from the original thread in which this worker was
        instantiated into this idle thread.
      * Calls the :meth:`QBetseeThreadPoolWorker.run` method of this worker.
      * Garbage collects this worker when this method returns.
    * Else, queues a request to subsequently run this worker in the next idle
      thread in this thread pool.

    Parameters
    ----------
    worker : QBetseeThreadPoolWorker
        Worker to be started in this thread pool.
    thread_pool : QThreadPoolOrNoneTypes
        Thread pool to start this worker in. Defaults to ``None``, in which
        case the singleton thread pool returned by the :func:`get_thread_pool`
        function is defaulted to.
    '''

    # Default this thread pool to the singleton thread pool if needed.
    if thread_pool is None:
        thread_pool = get_thread_pool()

    # Run this worker in this thread pool.
    thread_pool.start(worker)

# ....................{ HALTERS                           }....................
@type_check
def halt_workers(
    # Mandatory parameters.
    workers: IterableTypes,
    milliseconds: int,

    # Optional parameters.
    thread_pool: QThreadPoolOrNoneTypes = None,
) -> None:
    '''
    Wait no more than the passed number of milliseconds for all passed thread
    pool workers in the passed thread pool to gracefully stop and, if one or
    more workers fail to do so, non-gracefully terminate these workers
    immediately.

    Specifically, this function tests whether one or more thread pool workers
    are still working *and* fail to stop (gracefully or otherwise) after
    blocking the caller's thread for the passed number of milliseconds; if this
    is the case, each such worker is terminated by any means necessary.

    Parameters
    ----------
    workers : IterableTypes
        Workers to be halted in this thread pool.
    milliseconds : int
        Number of milliseconds (i.e., 10^-3 seconds) to block the current
        thread *before* non-gracefully halting these workers.
    thread_pool : QThreadPoolOrNoneTypes
        Thread pool to halt these workers in. Defaults to ``None``, in which
        case the singleton thread pool returned by the :func:`get_thread_pool`
        function is defaulted to.
    '''

    # Log this termination.
    logs.log_debug('Halting thread pool...')

    # If any such worker is *NOT* a worker, raise an exception.
    iterables.die_unless_items_instance_of(
        iterable=workers, cls=QBetseeThreadPoolWorker)

    # Default this thread pool to the singleton thread pool if needed.
    if thread_pool is None:
        thread_pool = get_thread_pool()

    # If *NO* thread pool workers are currently working, reduce to a noop.
    if not is_working(thread_pool):
        return
    # Else, one or more thread pool workers are still working.

    # Number of workers to be terminated.
    workers_num = len(workers)

    # Number of seconds to block.
    seconds = milliseconds * 10**-3

    # Log a non-fatal warning to inform the end user of this blocking event.
    logs.log_warning(
        'Waiting %g seconds for %d thread pool worker(s) to halt...',
        seconds, workers_num)

    # If these workers fail to stop after blocking for this duration of time...
    if not thread_pool.waitForDone(milliseconds):
        # Log this termination.
        logs.log_warning(
            'Forcing %d thread pool worker(s) to halt...', workers_num)

        # For each such worker, terminate this worker by any means necessary.
        for worker in workers:
            worker.halt()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulator** (i.e., :mod:`PySide2`-based object both displaying
*and* controlling the execution of simulation phases) functionality.
'''

#FIXME: When the application closure signal is emitted (e.g., from the
#QApplication.aboutToQuit() signal and/or QMainWindow.closeEvent() handler),
#the following logic should be performed (in order):
#
#1. In the QMainWindow.closeEvent() handler only:
#   * When the user attempts to close the application when one or more threads
#     are currently running, a warning dialog should be displayed to the user
#     confirming this action.

#FIXME: When the user attempts to run a dirty simulation (i.e., a simulation
#with unsaved changes), the GUI should prompt the user as to whether or not
#they would like to save those changes *BEFORE* running the simulation. In
#theory, we should be able to reuse existing "sim_conf" functionality to do so.

#FIXME: Note in a more appropriate docstring somewhere that the text overlaid
#onto the progress bar is only conditionally displayed depending on the current
#style associated with this bar. Specifically, the official documentation notes:
#
#    Note that whether or not the text is drawn is dependent on the style.
#    Currently CDE, CleanLooks, Motif, and Plastique draw the text. Mac, Windows
#    and WindowsXP style do not.
#
#For orthogonality with native applications, it's probably best to accept this
#constraint as is and intentionally avoid setting a misson-critical format on
#progress bars. Nonetheless, this *DOES* appear to be circumventable by
#manually overlaying a "QLabel" widget over the "QProgressBar" widget in
#question. For details, see the following StackOverflow answer (which, now that
#I peer closely at it, appears to be quite incorrect... but, something's better
#than nothing... maybe):
#    https://stackoverflow.com/a/28816650/2809027

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import Slot  #QCoreApplication, QObject, Signal
from betse.science.phase.phaseenum import SimPhaseKind
from betse.util.io.log import logs
from betse.util.type.text import strs
from betse.util.type.types import type_check
from betsee.gui.window.guimainwindow import QBetseeMainWindow
from betsee.gui.simtab.run.guisimrunact import QBetseeSimmerProactor
from betsee.gui.simtab.run.guisimrunenum import SimmerState, SimmerModelState
from betsee.gui.simtab.run.guisimrunstate import (
    SIM_PHASE_KIND_TO_NAME,
    SIMMER_STATES_IDLE,
    SIMMER_STATE_TO_PROACTOR_STATUS,
    SIMMER_STATE_TO_PROACTOR_SUBSTATUS,
    SIMMER_STATES_HALTING,
)
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ CLASSES                           }....................
class QBetseeSimmer(QBetseeControllerABC):
    '''
    High-level **simulator** (i.e., :mod:`PySide2`-based object both displaying
    *and* controlling the execution of simulation phases).

    Attributes (Private)
    ----------
    _progress_status_text_prior : str
        Most recent textual contents of the
        :attr:`QBetseeMainWindow.sim_run_player_status` label, preserved so as
        to permit this text to be transparently reused without localization
        concerns (particularly on repeatedly switching between the paused,
        stopped, and finished simulator states).

    Attributes (Private: Controllers)
    ----------
    _proactor : QBetseeSimmerProactor
        **Simulator proactor** (i.e., lower-level :mod:`PySide2`-based
        delegate controlling but *not* displaying the execution of simulation
        phases). In standard model-view-controller (MVC) parlance:

        * BETSE itself is the model (M) that runs simulation phases.
        * This proactor is the controller (C) for running simulation phases.
        * This parent object is the view (V) into running simulation phases.

    Attributes (Private: Widgets)
    ----------
    _action_toggle_work : QAction
        Alias of the :attr:`QBetseeMainWindow.action_sim_run_toggle_work`
        action.
    _action_stop_workers : QAction
        Alias of the :attr:`QBetseeMainWindow.action_sim_run_stop_workers`
        action.
    _player_toolbar : QFrame
        Alias of the :attr:`QBetseeMainWindow.sim_run_player_toolbar_frame`
        frame containing only the :class:`QToolBar` containing buttons for
        controlling the currently running simulation.
    _progress_bar : QProgressBar
        Alias of the :attr:`QBetseeMainWindow.sim_run_player_progress` widget.
    _progress_status : QLabel
        Alias of the :attr:`QBetseeMainWindow.sim_run_player_status` label,
        synopsizing the current state of this simulator.
    _progress_substatus : QLabel
        Alias of the :attr:`QBetseeMainWindow.sim_run_player_substatus` label,
        detailing the current state of this simulator.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this simulator.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Simulator proactor.
        self._proactor = QBetseeSimmerProactor(self)

        # Nullify all remaining instance variables for safety.
        self._action_toggle_work = None
        self._action_stop_workers = None
        self._player_toolbar = None
        self._progress_bar = None
        self._progress_status = None
        self._progress_substatus = None
        self._progress_status_text_prior = None


    @type_check
    def init(self, main_window: QBetseeMainWindow) -> None:
        '''
        Finalize this simulator's initialization, owned by the passed main
        window widget.

        This method connects all relevant signals and slots of *all* widgets
        (including the main window, top-level widgets of that window, and leaf
        widgets distributed throughout this application) whose internal state
        pertains to the high-level state of this simulator.

        To avoid circular references, this method is guaranteed to *not* retain
        references to this main window on returning. References to child
        widgets (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        '''

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Log this initialization.
        logs.log_debug('Sanitizing simulator state...')

        # Initialize all widgets concerning simulator state.
        self._init_widgets(main_window)

        # Connect all relevant signals and slots *AFTER* initializing these
        # widgets, as the former typically requires the latter.
        self._init_connections(main_window)


    @type_check
    def _init_widgets(self, main_window: QBetseeMainWindow) -> None:
        '''
        Create all widgets owned directly by this object *and* initialize all
        other widgets (*not* always owned by this object) concerning this
        simulator.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow`
            widget.
        '''

        # Classify variables of this main window required by this simulator.
        self._action_toggle_work  = main_window.action_sim_run_toggle_work
        self._action_stop_workers = main_window.action_sim_run_stop_work
        self._player_toolbar      = main_window.sim_run_player_toolbar_frame
        self._progress_bar        = main_window.sim_run_player_progress
        self._progress_status     = main_window.sim_run_player_status
        self._progress_substatus  = main_window.sim_run_player_substatus
        self._progress_substatus_group = (
            main_window.sim_run_player_substatus_group)


    @type_check
    def _init_connections(self, main_window: QBetseeMainWindow) -> None:
        '''
        Connect all relevant signals and slots of *all* widgets (including the
        main window, top-level widgets of that window, and leaf widgets
        distributed throughout this application) whose internal state pertains
        to the high-level state of this simulator.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow`
            widget.
        '''

        # Connect actions to corresponding slots of the proactor.
        self._action_toggle_work.triggered.connect(
            self._proactor.toggle_work)
        self._action_stop_workers.triggered.connect(
            self._proactor.stop_workers)

        # Connect widget signals to corresponding slots of this simulator.
        # Specifically:
        #
        # * When the proactor state is changed, update simulator widgets
        #   possibly dependent upon this state.
        # * When the progress bar is progressed, update simulator widgets
        #   possibly dependent upon this progress (e.g., progress substatus
        #   text interpolating the current and maximum progress bar values as
        #   the current and maximum time steps when modelling the
        #   initialization or simulation phase).
        self._proactor.state_changed.connect(
            self._sync_widgets_to_proactor_state)
        self._progress_bar.valueChanged.connect(
            self._sync_widgets_to_proactor_state_current)

        # Initialize the proactor *AFTER* establishing the prior signal-slot
        # connection, as the QBetseeSimmerProactor.init() method called here
        # internally changes the proactor state and hence requires that
        # simulator widgets be updated.
        self._proactor.init(
            main_window=main_window,
            progress_bar=self._progress_bar,
            progress_status=self._progress_status,
            progress_substatus=self._progress_substatus,
        )

    # ..................{ FINALIZERS                        }..................
    def halt_work(self) -> None:
        '''
        Schedule the currently running simulation phase if any for immediate
        and thus possibly non-graceful termination *or* silently reduce to a
        noop otherwise (i.e., if no simulation phase is currently running).

        Caveats
        ----------
        This method may induce data loss or corruption in simulation output.
        In theory, this should only occur in edge cases in which the current
        simulator worker fails to gracefully stop within a sensible window of
        time. In practice, this implies that this method should *only* be
        called when otherwise unavoidable (e.g., at application shutdown).

        See Also
        ----------
        :meth:`QBetseeSimmerProactor.halt_workers`
            Further details.
        '''

        # Transparently forward this request to our proactor.
        self._proactor.halt_workers()

    # ..................{ SLOTS ~ sync                      }..................
    @Slot(SimmerState, SimmerState)
    def _sync_widgets_to_proactor_state(
        self, state_new: SimmerState, state_old: SimmerState) -> None:
        '''
        Slot signalled on each transition of the proactor from the passed
        previous state into the passed current state.

        This slot synchronizes all widgets owned or otherwise controlled by
        this simulator to reflect this transition.

        Parameters
        ----------
        state_new : SimmerState
            Current state of the proactor.
        state_old : SimmerState
            Previous state of the proactor.
        '''

        # Log this slot.
        logs.log_debug('Synchronizing widgets to proactor state transition...')

        # If...
        if (
            # The proactor is transitioning to the idle state *OR*...
            state_new in SIMMER_STATES_IDLE or (
                # The proactor is transitioning to the modelling state *AND*...
                state_new is SimmerState.MODELLING and
                # The proactor is *NOT* already in the modelling state.
                state_old is not SimmerState.MODELLING
            )
        # ...then the proactor is either idle or beginning modelling. In either
        # case, reset the progress bar (i.e., prevent this bar from displaying
        # work progress).
        ):
            logs.log_debug('Resetting simulator progress bar...')
            self._progress_bar.reset()

        # Synchronize all widgets to reflect the current state of the proactor
        # *AFTER* possibly resetting the progress bar. The former explicitly
        # tests the latter condition.
        self._sync_widgets_to_proactor_state_current()


    @Slot()
    def _sync_widgets_to_proactor_state_current(self) -> None:
        '''
        Slot signalled on each transition of the proactor from its previous
        into its current state.

        This slot synchronizes all widgets owned or otherwise controlled by
        this simulator to reflect the current state of the proactor.
        '''

        # Log this slot.
        logs.log_debug('Synchronizing widgets to proactor state...')

        # Update the contents of the simulator toolbar *BEFORE* any other
        # updates, as the former takes precedence over the latter. Toolbar
        # buttons control the state of both the proactor and the currently
        # working phase (if any). Updating these buttons immediately on
        # entering this slot significantly increases the likelihood of these
        # buttons externally reflecting the internal state of both the proactor
        # and the currently working phase (if any).
        self._sync_toolbar()

        # Update all widgets depicting the proactor's progress.
        self._sync_progress()


    def _sync_toolbar(self) -> None:
        '''
        Update the contents of the **simulator toolbar** (i.e.,
        :class:`QToolBar` instance containing buttons controlling the currently
        running simulation).
        '''

        # Enable (in arbitrary order):
        #
        # * All widgets controlling the currently queued phase only if one or
        #   more phases are currently queued.
        # * All widgets halting the current worker only if some worker is
        #   currently working.
        #
        # To reduce the likelihood of accidental interaction with widgets
        # intended to be disabled, do so *BEFORE* subsequent slot logic.
        self._player_toolbar.setEnabled(self._proactor.is_queued)

        # Enable simulator starting, pausing, and resuming only if the
        # simulator is currently workable.
        self._action_toggle_work.setEnabled(self._proactor.is_workable)

        # Enable simulator pausing only if the simulator is currently running.
        self._action_toggle_work.setChecked(self._proactor.is_running)

        # Enable simulator stopping only if the simulator is currently working.
        # Note that:
        #
        # * Testing the "_is_running" property fails to suffice, as that
        #   property fails to yield "True" when the simulator is paused.
        # * While testing the "_is_worker" property alone could also
        #   theoretically suffice, doing so would desynchronize the UI from
        #   this simulator state; specifically, the button associated with this
        #   action would remain enabled for a non-deterministic window of time
        #   after the simulator is stopped. Why? Because the _stop_workers()
        #   slot sets this simulator state to stopped *BEFORE* the currently
        #   working worker successfully stops resulting in the "_is_worker"
        #   property yielding "False". Ergo, the simulator state takes
        #   precedence for UI purposes.
        self._action_stop_workers.setEnabled(self._proactor.is_working)


    def _sync_progress(self) -> None:
        '''
        Update the contents of all simulator widgets to reflect the current
        progress of the simulation subcommand being run by the proactor (if
        any).
        '''

        # Update the (sub)status of this simulator *AFTER* possibly updating
        # the state of the progress bar, as the former assumes the latter to
        # have already been performed.
        self._sync_progress_status()
        self._sync_progress_substatus()


    def _sync_progress_status(self) -> None:
        '''
        Update the text displayed by the :attr:`_progress_status` label,
        synopsizing the current state of this simulator.
        '''

        # Unformatted template synopsizing the current state of this simulator.
        status_text_template = SIMMER_STATE_TO_PROACTOR_STATUS[
            self._proactor.state]

        # Text synopsizing the type of simulation phase run by the current
        # simulator worker if any *OR* an arbitrary placeholder otherwise. In
        # the latter case, this text is guaranteed to *NOT* be interpolated by
        # this template and is thus safely ignorable.
        phase_type_name = (
            SIM_PHASE_KIND_TO_NAME[self._proactor.worker.phase.kind]
            if self._proactor.is_worker else
            'the nameless that shall not be named')

        # If no prior progress status exists, this is the first call to this
        # method. In this case, the current state of the proactor is guaranteed
        # to either be the queued or unqueued states; in either case, the
        # progress status template under this state ignores the prior progress
        # status. This implies that this status could technically be
        # initialized to any string with no ill consequences. For simplicity,
        # default this status to the current progress status.
        if self._progress_status_text_prior is None:
            self._progress_status_text_prior = self._progress_status.text()

        # Unconditionally format this text with all possible format specifiers
        # expected by all possible instances of this text. (Note that Python
        # ignores format specifiers *NOT* expected by this exact text.)
        status_text = status_text_template.format(
            phase_type=phase_type_name,
            status_prior=self._progress_status_text_prior,
        )

        # If the proactor is *NOT* currently in a halting state (i.e., is
        # either running, queued, or unqueued)...
        if self._proactor.state not in SIMMER_STATES_HALTING:
            # Preserve this progress status for subsequent use when halting.
            self._progress_status_text_prior = status_text

            # Enable this text to be interpolated into arbitrary character
            # positions by (in arbitrary order):
            #
            # * Lowercasing the first character of this text.
            # * Stripping the trailing ellipses from this text if present.
            self._progress_status_text_prior = (
                strs.lowercase_char_first(
                    strs.remove_suffix_if_found(
                        text=self._progress_status_text_prior, suffix='...')))
        # Else, the proactor is currently in a halting state (i.e., is either
        # paused, stopped, or finished). In this case, preserve the prior
        # progress status for subsequent reuse when switching between halting
        # states. If this edge case were *NOT* explicitly detected here, then
        # switching from the stopped to finished states would, for example,
        # result in progress status resembling "Finished stopped modelling...".

        # Set the text of the label displaying this synopsis to this text.
        self._progress_status.setText(status_text)


    def _sync_progress_substatus(self) -> None:
        '''
        Update the text displayed by the :attr:`_progress_substatus` label,
        detailing the current state of this simulator.
        '''

        # Arbitrary object detailing the current state of this simulator, whose
        # type is specific to this state.
        substatus_value = SIMMER_STATE_TO_PROACTOR_SUBSTATUS[
            self._proactor.state]

        # Unformatted template detailing the current state of this simulator,
        # conditionally defined below in a state-dependent manner.
        substatus_text_template = None

        # If the proactor is currently modelling...
        if self._proactor.state is SimmerState.MODELLING:
            # Type of simulaton phase currently being modelled.
            phase_kind = self._proactor.worker.phase.kind

            # If the proactor is currently modelling the seed phase, silently
            # reduce to a noop. Details of modelling the seed phase are already
            # displayed by the slot connected to the signal emitted by the
            # SimCallbacksSignaller.progress_stated() of the current simulator
            # worker in a push-driven manner.
            if phase_kind is SimPhaseKind.SEED:
                return
            # Else, the proactor is currently modelling either the
            # initialization or simulation phases.

            # Type of simulator modelling currently being performed.
            model_state = None

            # If the progress bar is currently in the reset state, then the
            # current proactor worker performing this modelling has yet to
            # model any time steps of this simulation and hence is in the
            # pre-processing state.
            #
            # Note that the _sync_widgets_to_proactor_state() method explicitly
            # resets the progress bar on transitioning into the modelling
            # state, ensuring sanity here.
            if self._progress_bar.is_reset:
                model_state = SimmerModelState.PREPARING
            # Else, this worker has modelled one or more time steps.
            #
            # If the progress bar is currently in the finished state, then this
            # this worker has modelled all time steps of this simulation and
            # hence is in the post-processing state.
            elif self._progress_bar.is_done:
                model_state = SimmerModelState.FINISHING
            # Else, this worker has modelled one or more but *NOT* all time
            # steps and hence is in the in-processing state.
            else:
                model_state = SimmerModelState.MODELLING

            # Dictionary mapping from from each type of simulator modelling
            # state to an unformatted string template detailing the current
            # action being performed in that state.
            model_state_to_substatus_text_template = substatus_value[
                phase_kind]

            # Unformatted string template detailing the current action.
            substatus_text_template = model_state_to_substatus_text_template[
                model_state]
        # Else if the proactor is currently exporting...
        elif self._proactor.state is SimmerState.EXPORTING:
            #FIXME: Refactor this to actually do something *AFTER* resolving
            #the lower-level concerns in the BETSE codebase. Notably, see
            #commentary in the SimPipesExport.export() method for details.

            # Silently reduce to a noop. Details of exporting *ANY* phase are
            # already displayed by the slot connected to the signal emitted by
            # the SimCallbacksSignaller.progress_stated() of the current
            # simulator worker in a push-driven manner... which is awful.
            return
        # Else, the proactor is neither modelling nor exporting. In this case,
        # the details of the proactor's current action reduce to a string.
        else:
            substatus_text_template = substatus_value

        # Text detailing the prior state of this simulator.
        substatus_text_prior = self._progress_substatus.text()

        # Metadata synopsizing the current state of this simulator.
        proactor_metadata = self._proactor.phaser.get_metadata()

        # Unconditionally format this text with all possible format specifiers
        # expected by all possible instances of this text. (Note that Python
        # ignores format specifiers *NOT* expected by this exact text.)
        substatus_text = substatus_text_template.format(
            progress_current=self._progress_bar.value(),
            progress_maximum=self._progress_bar.maximum(),
            queued_modelling=proactor_metadata.phases_queued_modelling_count,
            queued_exporting=proactor_metadata.phases_queued_exporting_count,
            substatus_prior=substatus_text_prior,
        )

        # Set the text of the label displaying these details to this text.
        self._progress_substatus.setText(substatus_text)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Application-wide **focus** (i.e., interactive keyboard input focus received by
the current application widget) functionality.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication
from PySide2.QtWidgets import QApplication, QWidget
from betsee.guiexception import BetseePySideFocusException
from betsee.util.type.guitype import QWidgetOrNoneTypes

# ....................{ EXCEPTIONS                         }....................
def die_unless_widget_focused() -> None:
    '''
    Raise an exception if *no* application widget currently has the interactive
    keyboard input focus.

    Raises
    ----------
    BetseePySideFocusException
        If *no* widget is currently focused.

    See Also
    ----------
    :func:`is_widget_focused`
        Further details.
    '''

    if not is_widget_focused():
        raise BetseePySideFocusException(QCoreApplication.translate(
            'die_unless_widget_focused', 'No widget currently focused.'))

# ....................{ TESTERS                            }....................
def is_widget_focused() -> bool:
    '''
    ``True`` only if some application widget currently has the interactive
    keyboard input focus.
    '''

    # Don't ask. Don't tell.
    return QApplication.focusWidget() == 0

# ....................{ GETTERS                            }....................
def get_widget_focused() -> QWidget:
    '''
    Application widget that currently has the interactive keyboard input focus
    if any *or* raise an exception otherwise.

    Returns
    ----------
    QWidget
        Currently focused widget.

    Raises
    ----------
    BetseePySideFocusException
        If *no* widget is currently focused.
    '''

    # If no widget is currently focused, raise an exception.
    die_unless_widget_focused()

    # Else, some widget is currently focused. Return this widget.
    return QApplication.focusWidget()


def get_widget_focused_or_none() -> QWidgetOrNoneTypes:
    '''
    Application widget that currently has the interactive keyboard input focus
    if any *or* ``None`` otherwise.

    Returns
    ----------
    QWidgetOrNoneTypes
        Currently focused widget if any *or* ``None`` otherwise.
    '''

    # Currently focused widget if any or 0 otherwise.
    widget_focused = QApplication.focusWidget()

    # Return None if no widget is focused or this widget otherwise.
    return None if widget_focused == 0 else widget_focused



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.

'''
High-level application initialization common to both the CLI and GUI.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To defer heavyweight and possibly circular imports, the top-level of
# this module may import *ONLY* from standard Python packages. All imports from
# application and third-party packages should be deferred to their point of use.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ INITIALIZERS                       }....................
def reinit() -> None:
    '''
    (Re-)initialize this application -- but *not* mandatory third-party
    dependencies of this application, which requires external resources (e.g.,
    command-line options, configuration files) to be parsed.

    Specifically, this function (in order):

    #. Initializes all lower-level BETSE logic by calling the
       :func:`betse.ignition.init` function.
    #. Validates but does *not* initialize mandatory third-party dependencies of
       this application, which must be initialized independently by the
       :func:`betsee.lib.guilib.init` function.
    #. Validates the active Python interpreter to support multithreading.

    Design
    ----------
    To support caller-specific error handling, this function is intended to be
    called immediately *after* this application begins catching otherwise
    uncaught exceptions.

    Whereas BETSE is intended to be both run non-interactively from the
    command-line and imported interactively from Python REPLs (e.g., Jupyter),
    BETSEE is intended to only be run non-interactively from desktop application
    launchers. In the former case, BETSE detects and ignores attempts to
    re-initialize itself in the same application process. In the latter case, no
    re-initialization is expected, detected, or ignored.

    See Also
    ----------
    :func:`betsee.lib.libs.reinit`
        Function (re)-initializing all mandatory third-party dependencies.
    '''

    # Defer heavyweight and possibly circular imports.
    from betse import ignition as betse_ignition
    from betsee.lib import guilib

    # Initialize all lower-level BETSE logic *BEFORE* any higher-level BETSEE
    # logic requiring the former. See the betse.guicli.BetseCLI._ignite_app()
    # method for details on why the betse_ignition.reinit() rather than
    # betse_ignition.init() function is called here.
    betse_ignition.reinit()

    # Validate mandatory dependencies *AFTER* initializing BETSE.
    #
    # These dependencies are intentionally *NOT* initialized here (e.g., by
    # calling guilib.init()), as doing so requires the logging configuration to
    # have been finalized (e.g., by parsing CLI options), which has yet to
    # occur this early in the application lifecycle.
    guilib.die_unless_runtime_mandatory_all()

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # WARNING: Avoid logic below that attempts to perform logging. (See above.)
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **multithreading** (i.e., platform-portable, pure-Qt,
:class:`QThread`-based parallelization implemented external to Python and hence
Python's restrictive Global Interpreter Lock (GIL)) facilities.
'''

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QCoreApplication, QEventLoop, QThread
from betse.util.io.log import logs
from betse.util.py import pythread
from betse.util.type.types import type_check
from betsee.guiexception import BetseePySideThreadException
from betsee.util.type.guitype import QThreadOrNoneTypes

# ....................{ GLOBALS                           }....................
_THREAD_MAIN_ID = None
'''
Arbitrary Qt-specific integer uniquely identifying the main thread in the
active Python interpreter if the currently installed version of PySide2 exposes
the :meth:`QThread.currentThreadId` method yielding this integer *or* the
Python-specific ID of this thread's object wrapper instead.

Caveats
----------
This integer is non-portable and hence should *only* be embedded in low-level
debugging messages. Moreover, this integer is ``None`` until the :func:`init`
function is called. In other words, you probably want to try something else.
'''

# ....................{ EXCEPTIONS                        }....................
def die_if_thread_current_main() -> None:
    '''
    Raise an exception if the current thread is the main thread in the active
    Python interpreter.

    This function should typically be called for safety by callers attempting
    to subsequently block the current thread (e.g., by calling the
    :func:`wait_for_events_if_none` function). The main thread administers the
    Qt-specific GUI event loop and hence should *never* be blocked for any
    duration of time whatsoever.

    Raises
    ----------
    BetseePySideThreadException
        If this function is called from the main thread.

    See Also
    ----------
    :func:`is_thread_current_main`
        Further details.
    '''

    if is_thread_current_main():
        raise BetseePySideThreadException(QCoreApplication.translate(
            'die_if_thread_current_main',
            'Operation prohibited in main thread.'))

# ....................{ EXCEPTIONS ~ private              }....................
def _die_unless_thread_main_id() -> None:
    '''
    Raise an exception unless a prior call to the :func:`init` function has
    defined the :data:`_THREAD_MAIN_ID` global.

    Raises
    ----------
    BetseePySideThreadException
        If the :func:`init` function has yet to be called.
    '''

    if _THREAD_MAIN_ID is None:
        raise BetseePySideThreadException(QCoreApplication.translate(
            '_die_unless_thread_main_id',
            '"_THREAD_MAIN_ID" undefined (e.g., '
            'due to the init() function having yet to be called).'))

# ....................{ TESTERS                           }....................
@type_check
def is_thread_current_main() -> bool:
    '''
    ``True`` only if the current thread is the main thread in the active Python
    interpreter.
    '''

    # Raise an exception unless the "_THREAD_MAIN_ID" global is defined.
    _die_unless_thread_main_id()

    # Return true only if this thread's identifier is that of the main thread.
    return get_thread_current_id() == _THREAD_MAIN_ID


@type_check
def should_halt_thread_work(thread: QThreadOrNoneTypes = None) -> bool:
    '''
    ``True`` only if some object has requested that the passed thread of
    execution voluntarily halt *all* tasks (e.g., overriden :meth:`QThread.run`
    method) and workers (e.g., subclassed :class:`QBetseeThreadWorkerABC`
    object) currently running in this thread.

    This function is intended to be voluntarily called by tasks and workers.

    Caveats
    ----------
    One parent thread may run an arbitrary number of child workers. However,
    the request to halt tested by this function is a low-level condition
    applying to a single thread rather than a single worker. Hence, if this
    function returns ``False``, *all* tasks and workers currently running in
    this thread are expected to gracefully halt all work being performed and
    then terminate.

    For fine-grained control over worker lifecycles, external callers are
    strongly advised to signal each such worker to gracefully halt (e.g., by
    emitting a signal connected to the :meth:`QBetseeThreadWorkerABC.stop`
    slot) rather than requesting that the thread running those workers halt
    (e.g., by calling the :meth:`QThread.requestInterruption` method).
    Nonetheless, workers are expected to respect both types of requests.

    For convenience, note that the
    :meth:`QBetseeThreadWorkerABC._halt_work_if_requested` method implicitly
    respects both types of requests. Ergo, :class:`QBetseeThreadWorkerABC`
    subclasses need *not* and should *not* explicitly call this function.

    Parameters
    ----------
    thread : QThreadOrNoneTypes
        Thread to request to be halted. Defaults to ``None``, in which case the
        current thread of execution is requested to be halted.

    Returns
    ----------
    bool
        ``True`` only if some object has requested this thread to voluntarily
        halt *all* tasks and workers currently running in this thread.

    See Also
    ----------
    :func:`halt_thread_work`
        Companion function requesting this thread to voluntarily halt these
        tasks and workers.
    '''

    # Default this thread if unpassed to the current thread of execution.
    if thread is None:
        thread = get_thread_current()

    # Return true only if this thread has been requested to be halted.
    return thread.isInterruptionRequested()

# ....................{ GETTERS ~ current                 }....................
def get_thread_current() -> QThread:
    '''
    Wrapper encapsulating the **current thread** (i.e., Qt-specific thread of
    execution responsible for the call to this function).
    '''

    return QThread.currentThread()


# Placeholder function definition to avoid spurious linter errors. For
# efficiency, this frequently called function is conditionally defined by the
# init() function at application initialization time.
def get_thread_current_id() -> int:
    pass


def get_thread_current_process_name() -> str:
    '''
    Name of the low-level kernel-level process associated with the current
    thread, equivalent to the Qt-specific name of this object.

    See Also
    ----------
    :meth:`betsee.util.thread.guithreadcls.QBetseeWorkerThread.process_name`
        Further details.
    '''

    # Current thread.
    thread = get_thread_current()

    # Return this thread's process name.
    return thread.objectName()

# ....................{ GETTERS ~ current : event         }....................
#FIXME: This submodule is increasingly awkward, largely due to the unctuous
#requirement that a "thread" reference always be passed whenever an
#"event_dispatcher" reference is required. Instead, do the following:
#
#* Define a new "ThreadEventDispatcherWrapper" class subclassing "object"
#  rather than "QObject" in this submodule.
#* Define the following instance variables of this class:
#  * "thread".
#  * "event_dispatcher".
#* Redefine the process_thread_events() and wait_for_events_if_none() functions into
#  methods of this class operating upon these variables.
#* Redefine the get_event_dispatcher_current() to return an instance of the
#  "ThreadEventDispatcherWrapper" class.
def get_thread_event_dispatcher_current() -> tuple:
    '''
    2-tuple ``(thread, event_dispatcher)`` encapsulating both the thread *and*
    event dispatcher for the current thread of execution if this thread has an
    event dispatcher *or* raise an exception otherwise (i.e., if this thread
    has no event dispatcher).

    Design
    ----------
    This getter intentionally returns both the thread *and* event dispatcher
    for the current thread of execution rather than merely the latter. Why?
    Because it appears infeasible to pass around arbitrary instances of the
    :class:`QAbstractEventDispatcher` superclass. For unknown reasons, some
    combination of Python and/or Qt aggressively collect an event dispatcher in
    the absence of a strong reference to the thread of that event dispatcher.
    Ergo, the two *must* be passed around as a single, cohesive unit.

    Returns
    ----------
    (QThread, QAbstractEventDispatcher)
        2-tuple ``(thread, event_dispatcher)``, where:

        * ``thread`` is the Qt-based wrapper encapsulating the current thread.
        * ``event_dispatcher`` is the Qt-based wrapper for the event dispatcher
          of the current thread.

    Raises
    ----------
    BetseePySideThreadException
        If this thread has *no* event dispatcher.
    '''

    # Current thread and event dispatcher of this thread (if any).
    thread = QThread.currentThread()
    event_dispatcher = thread.eventDispatcher()

    # If this thread has *NO* event dispatcher, raise an exception.
    if not event_dispatcher:
        raise BetseePySideThreadException(QCoreApplication.translate(
            'get_thread_event_dispatcher_current',
            'No event dispatcher running for thread "{}".'.format(
                thread.objectName())))

    # Return this thread and event dispatcher as a single, cohesive unit.
    return thread, event_dispatcher

# ....................{ LOGGERS ~ current                 }....................
@type_check
def log_debug_thread_current(message: str, *args) -> None:
    '''
    Log the passed debug message from any arbitrary thread with the root
    logger, formatted with the passed ``%``-style positional arguments.

    Parameters
    ----------
    message : str
        Human-readable message to be logged.

    All remaining positional parameters are conditionally interpolated as
    ``%``-style format specifiers into this message by logging facilities.

    See Also
    ----------
    :func:`log_debug_thread_main`
        Faster analogue intended to be called *only* when the current thread is
        the main thread.
    '''

    # Log this message at the debug level. See log_main_thread_debug().
    logs.log_debug(
        message + ' <from thread "%d">', *args, get_thread_current_id())


@type_check
def log_warning_thread_current(message: str, *args) -> None:
    '''
    Log the passed warning message from any arbitrary thread with the root
    logger, formatted with the passed ``%``-style positional arguments.

    Parameters
    ----------
    message : str
        Human-readable message to be logged.

    All remaining positional parameters are conditionally interpolated as
    ``%``-style format specifiers into this message by logging facilities.
    '''

    # Log this message at the warning level. See log_main_thread_debug().
    logs.log_warning(
        message + ' <from thread "%d">', *args, get_thread_current_id())

# ....................{ LOGGERS ~ main                    }....................
@type_check
def log_debug_thread_main(message: str, *args) -> None:
    '''
    Log the passed debug message from the main thread with the root logger,
    formatted with the passed ``%``-style positional arguments.

    Caveats
    ----------
    This function is intended to be called *only* from the main thread. For
    efficiency, this function does *not* validate this to be the case.

    Parameters
    ----------
    message : str
        Human-readable message to be logged.

    All remaining positional parameters are conditionally interpolated as
    ``%``-style format specifiers into this message by logging facilities.

    See Also
    ----------
    :func:`log_debug_thread_current`
        Slower analogue intended to be called from *any* thread.
    '''

    # Log this message at the debug level. To ensure that "%"-style format
    # specifiers embedded in this message are properly deferred until logging
    # time rather than performed now, the desired log message is created with
    # simple string concatenation rather than deferred until logging time with
    # "%"-style format specifiers, which would conflict with the former.
    logs.log_debug(
        message + ' <from main thread "%d">', *args, _THREAD_MAIN_ID)

# ....................{ HALTERS                           }....................
@type_check
def halt_thread_work(thread: QThreadOrNoneTypes = None) -> bool:
    '''
    Requested that the passed thread of execution voluntarily halt *all* tasks
    (e.g., overriden :meth:`QThread.run` method) and workers (e.g., subclassed
    :class:`QBetseeThreadWorkerABC` object) currently running in this thread.

    This function is intended to be called from objects in other threads.

    Caveats
    ----------
    After this function is called, this request will be unconditionally
    preserved in this thread until this thread is manually halted and restarted
    (e.g., by calling the :meth:`QBetseeWorkerThread.halt` and
    :meth:`QBetseeWorkerThread.start` methods in that order). Sadly, doing so:

    * Terminates all tasks and workers currently running in this thread,
      typically in a non-graceful manner resulting in in-memory or on-disk data
      corruption.
    * Discards all pending events currently queued with this thread --
      including both outgoing signals emitted by *and* incoming slots signalled
      on any workers still running in this thread.

    Hence, due to long-standing deficiencies in the Qt API, this request
    *cannot* be gracefully "undone." Attempting to do so *always* runs a risk
    of non-gracefully terminating running and pending work. The only
    alternative to this extremely concerning caveat is to signal each such
    worker to gracefully halt (e.g., by emitting a signal connected to the
    :meth:`QBetseeThreadWorkerABC.stop` slot) rather than calling this method.

    **You have been warned.** There be vicious vipers scuttling about here.

    Parameters
    ----------
    thread : QThreadOrNoneTypes
        Thread to request to be halted. Defaults to ``None``, in which case the
        current thread of execution is requested to be halted.

    See Also
    ----------
    :func:`should_halt_thread_work`
        Further details.
    https://forum.qt.io/topic/43067/qthread-requestinterruption-cannot-be-undone
        Prominent Qt forum discussion on this lamentable topic.
    '''

    # Default this thread if unpassed to the current thread of execution.
    if thread is None:
        thread = get_thread_current()

    # Request this thread's work to be halted.
    return thread.requestInterruption()

# ....................{ EVENTS ~ current                  }....................
@type_check
def process_thread_current_events() -> None:
    '''
    Process all pending events queued with the current thread.

    This function unconditionally processes *all* such events, including both
    outgoing signals emitted by *and* incoming slots signalled on all workers
    currently running in this thread.

    Caveats
    ----------
    Avoid manually calling the static :func:`QCoreApplication.processEvents`
    function to process events for the current thread of execution. If this
    thread has *no* event dispatcher, that function silently and hence unsafely
    ignores all pending :class:`DeferredDelete` events currently queued for
    this thread. Since numerous standard widgets (e.g., :class:`QToolTip`)
    require these events to be processed in a timely manner, ignoring these
    events fundamentally disrupts widget sanity.
    '''

    # Current thread and event dispatcher of this thread.
    thread, event_dispatcher = get_thread_event_dispatcher_current()

    # Process all pending events currently queued with this dispatcher.
    event_dispatcher.processEvents(QEventLoop.AllEvents)


@type_check
def wait_for_thread_current_event_if_none() -> None:
    '''
    Indefinitely block the current thread until *another* thread posts a new
    event to the current thread if *no* pending events are queued with the
    current thread *or* safely return immediately otherwise (i.e., if one or
    more pending events are currently queued with this thread).

    This function is intended to be called *only* from the slots of workers
    currently running in the thread of execution associated with the passed
    event dispatcher. In this case, this function call gracefully returns when
    another object in another thread of execution emits a signal connected to a
    slot of such a worker.

    Caveats
    ----------
    **Never call this function from an object owned by the main event-handling
    thread.** Doing so effectively blocks the application GUI managed by this
    thread. Instead, only call this function from a worker guaranteed to be
    running in this thread.
    '''

    # If the current thread is the main thread, raise an exception. The main
    # thread should *NEVER* be blocked for any period of time whatsoever.
    die_if_thread_current_main()

    # Current thread and event dispatcher of this thread.
    thread, event_dispatcher = get_thread_event_dispatcher_current()

    # Process all pending events currently queued with this dispatcher.
    event_dispatcher.processEvents(QEventLoop.WaitForMoreEvents)

# ....................{ INITIALIZERS                      }....................
def init() -> None:
    '''
    Initialize this submodule.

    Caveats
    ----------
    **This function is intended to be called from the main thread.** If this is
    *not* the case, the :data:`_THREAD_MAIN_ID` global will be improperly
    initialized.

    Raises
    ----------
    BetsePyException
        If the active Python interpreter does *not* support multithreading.
    '''

    # Submodule attributes at global scope (re)defined below. Yes, you declare
    # a function at global scope from within another function via the same
    # "global" declarator; failing to do so results in the former function
    # being declared as a locally scoped closure, instead. See also:
    #     https://stackoverflow.com/a/27930100/2809027
    global _THREAD_MAIN_ID, get_thread_current_id

    # Log this initialization.
    logs.log_debug('Initializing multithreading facilities...')

    # If the active Python interpreter does *NOT* support multithreading, raise
    # an exception *AFTER* BETSE is initialized.
    #
    # While Qt-based multithreading primitives are typically preferable to
    # Python-based multithreading primitives, the latter have their occasional
    # use. More critically, the lack of latter typically implies this
    # interpreter to be in insufficiently sane for our purposes.
    pythread.die_unless_threadable()

    # Conditionally define the frequently called get_thread_current_id() getter
    # by deciding whether PySide2 provides a binding for the static
    # QThread.currentThreadId() method on the current system and adopting the
    # best implementation available.
    #
    # Although Qt guarantees this method to exist, PySide2 bindings make no
    # such guarantees. For unknown reasons, all recent releases of PySide2 have
    # reliably failed to provide a binding for this method. Yes, this is a bug.
    try:
        QThread.currentThreadId
    # If PySide2 fails to expose this method, fallback to an implementation
    # returning the Python-specific ID of the object wrapper for this thread.
    except AttributeError:
        # Log this mishap as a non-fatal warning. What can we do, right?
        logs.log_warning(
            'QThread.currentThreadId() undefined; '
            'falling back to thread wrapper ID.')

        # Define this getter in the most hideous way possible.
        def get_thread_current_id() -> int:
            return id(get_thread_current())
    # Else, PySide2 exposes this method. In this case, adopt the preferred
    # implementation of simply deferring to this method.
    else:
        def get_thread_current_id() -> int:
            return QThread.currentThreadId()

    # Unconditionally document this getter regardless of implementation.
    get_thread_current_id.__doc__ = '''
    Arbitrary Qt-specific integer uniquely identifying the current thread in
    the active Python interpreter if the currently installed version of PySide2
    exposes the :meth:`QThread.currentThreadId` method yielding this integer
    *or* the Python-specific ID of this thread's object wrapper instead.

    Caveats
    ----------
    This integer is non-portable and hence should *only* be embedded in
    low-level debugging messages.

    Note that there exists no general-purpose :func:`get_thread_id` getter
    returning the unique identifier for an arbitrary thread, as Qt defines no
    corresponding functionality for doing so.

    See Also
    ----------
    :func:`get_thread_current`
        Further details.
    '''

    # Immediately identify the main thread (assumed without proof to be the
    # current thread) with the same function.
    _THREAD_MAIN_ID = get_thread_current_id()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
General-purpose :mod:`QProgressBar` subclasses.
'''

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import Slot  # Signal,
from PySide2.QtWidgets import QProgressBar
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ SUBCLASSES                        }....................
class QBetseeProgressBar(QBetseeObjectMixin, QProgressBar):
    '''
    :mod:`QProgressBar`-based widget exposing additional caller-friendly slots.

    This widget augments the stock :class:`QProgressBar` class with high-level
    properties simplifying usage *and* high-level slots combining the utility
    of various lower-level slots (e.g., :meth:`set_range_and_value_minimum`).
    '''

    # # ..................{ INITIALIZERS                      }..................
    # def __init__(self, *args, **kwargs) -> None:
    #
    #     # Initialize our superclass with all passed parameters.
    #     super().__init__(*args, **kwargs)

    # ..................{ PROPERTIES                        }..................
    @property
    def is_done(self) -> bool:
        '''
        ``True`` only if this progress bar is currently in the **finished
        state** (i.e., if the current value of this progress bar is equal to
        the maximum value previously passed to the :meth:`setRange` method).
        '''

        # Surprisingly, this actually makes sense.
        return self.value() == self.maximum()


    @property
    def is_reset(self) -> bool:
        '''
        ``True`` only if this progress bar is currently in the **reset state**
        (i.e., if either the :meth:`setValue` method has yet to be called *or*
        the :meth:`reset` method has been called more recently than the
        :meth:`setValue` method).

        Equivalently, this property returns ``True`` only if this progress bar
        has no current progress value. Note that:

        * All progress bars are initially reset by default.
        * The **undetermined state** (i.e., when the range of this progress
          bar is ``[0, 0]``) takes precedence over this reset state. A progress
          bar that is currently undetermined is *not* reset.
        '''

        #FIXME: While unlikely to ever crop in our progress bar use cases,
        #consider the "INT_MIN" case as well. What is "INT_MIN", exactly?
        #Sadly, extensive grepping about suggests that neither PyQt5 nor
        #PySide2 expose this constant to Python code. Ergo, checking that
        #appears to infeasible. The alternative would be to:
        #
        #* Define a new "_is_reset" boolean instance variable.
        #* Default this boolean to "False" in the __init__() method.
        #* Override the reset() method to enable this boolean.
        #* Override the setValue() method to disable this boolean.

        # Yes, this is insanity. Nonetheless, this condition corresponds
        # exactly to the single-line C++ action performed by the default
        # QProgressBar.reset() implementation:
        #
        #    void QProgressBar::reset()
        #    {
        #        Q_D(QProgressBar);
        #        d->value = d->minimum - 1;
        #        if (d->minimum == INT_MIN)
        #            d->value = INT_MIN;
        #        repaint();
        #    }
        return self.value() == self.minimum() - 1


    @property
    def is_undetermined(self) -> bool:
        '''
        ``True`` only if this progress bar is currently in the **undetermined
        state** (i.e., if the :meth:`set_range_undetermined` method has been
        called more recently than any other range-setting method).

        Equivalently, this property returns ``True`` only if this progress bar
        allows exactly one possible progress value (i.e., 0), which Qt
        typically portrays as a busy indicator.
        '''

        # Insanity, thy name is the Qt API.
        return (
            self.maximum() == 0 and  # ...short-circuit us up the bomb!
            self.minimum() == 0
        )

    # ..................{ SLOTS                             }..................
    @Slot(int, int)
    def set_range_and_value_minimum(self, minimum: int, maximum: int) -> None:
        '''
        Sets the range of this progress bar to the passed range *and* sets the
        value of this progress bar to the minimum value of this range.

        This high-level slot is intended to serve as a drop-in replacement for
        the lower-level :meth:`setRange`. The latter only sets the range of
        this progress bar to the passed range, requiring callers to explicitly
        call the stock :meth:`setValue` slot with the minimum value of this
        range to perform the equivalent of this slot. Since callers typically
        perform both on setting the range, this slot conjoins these two slots
        into one slot for caller convenience.

        Caveats
        ----------
        Note that, if:

        * The passed minimum and maximum are both 0, this progress bar is set
          to an undetermined state.
        * The passed maximum is less than the passed minimum, only the passed
          minimum is set; the passed maximum is ignored.
        * The current value of this progress bar resides outside this range,
          this progress bar is implicitly reset by internally calling the
          :meth:`reset` method.

        Parameters
        ----------
        minimum : int
            Minimum value to constrain the values of this progress bar to.
        maximum : int
            Maximum value to constrain the values of this progress bar to.
        '''

        # Set the current range to the passed range.
        self.setRange(minimum, maximum)

        # Set the current value to the minimum of this range.
        self.setValue(minimum)

    # ..................{ METHODS                           }..................
    def set_range_undetermined(self) -> None:
        '''
        Change this progress bar to the **undetermined state** (i.e., the state
        such that this progress bar allows exactly one possible progress value
        of 0 rather than a range of such values).

        Qt typically portrays this state as a busy indicator.
        '''

        # Set the current range to an arbitrary pair of magic numbers. *sigh*
        self.setRange(0, 0)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulator phaser** (i.e., :mod:`PySide2`-based object containing
each simulator phase controller on behalf of higher-level parent controllers)
functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.science.phase.phaseenum import SimPhaseKind
from betse.util.io.log import logs
from betse.util.type.iterable import tuples
from betse.util.type.types import type_check, CallableTypes, QueueType
from betsee.gui.window.guimainwindow import QBetseeMainWindow
from betsee.gui.simtab.run.phase.guisimrunphase import QBetseeSimmerPhase
from betsee.gui.simtab.run.work.guisimrunwork import QBetseeSimmerPhaseWorker
from betsee.gui.simtab.run.work.guisimrunworkenum import SimmerPhaseSubkind
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC
from collections import deque

# ....................{ TYPES                             }....................
SimmerProactorMetadata = tuples.make_named_subclass(
    class_name='SimmerProactorMetadata',
    item_names=(
       'phases_queued_modelling_count',
       'phases_queued_exporting_count',
    ),
    doc='''
    Named tuple created and returned by the
    :meth:`QBetseeSimmerProactor.get_metadata` method, aggregating metadata
    synopsizing the current state of the simulator proactor.

    Attributes
    ----------
    phases_queued_modelling_count : int
        Number of simulator phases currently queued for modelling.
    phases_queued_exporting_count : int
        Number of simulator phases currently queued for exporting.
    '''
)

# ....................{ CLASSES                           }....................
class QBetseeSimmerPhaser(QBetseeControllerABC):
    '''
    High-level **simulator proactor phaser** (i.e., :mod:`PySide2`-based
    object containing each simulator phase controller on behalf of
    higher-level parent controllers).

    Attributes (Public)
    ----------
    PHASES : SequenceTypes
        Immutable sequence of all simulator phase controllers (e.g.,
        :attr:`_phase_seed`), needed for iteration over these controllers. For
        sanity, these phases are ordered is simulation order such that:

        * The seed phase is listed *before* the initialization phase.
        * The initialization phase is listed *before* the simulation phase.

    Attributes (Private)
    ----------
    _phase_seed : QBetseeSimmerPhase
        Controller for all simulator widgets pertaining to the seed phase.
    _phase_init : QBetseeSimmerPhase
        Controller for all simulator widgets pertaining to the initialization
        phase.
    _phase_sim : QBetseeSimmerPhase
        Controller for all simulator widgets pertaining to the simulation
        phase.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this simulator proactor phaser.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Simulator phase controllers.
        self._phase_seed = QBetseeSimmerPhase(self)
        self._phase_init = QBetseeSimmerPhase(self)
        self._phase_sim  = QBetseeSimmerPhase(self)

        # Sequence of all simulator phases, intentionally listed in simulation
        # order to ensure sane iterability.
        self.PHASES = (self._phase_seed, self._phase_init, self._phase_sim)


    @type_check
    def init(
        self,
        main_window: QBetseeMainWindow,
        set_state_from_phase: CallableTypes,
    ) -> None:
        '''
        Finalize this phaser's initialization, owned by the passed main window
        widget.

        This method connects all relevant signals and slots of *all* widgets
        (including the main window, top-level widgets of that window, and leaf
        widgets distributed throughout this application) whose internal state
        pertains to the high-level state of this simulator.

        To avoid circular references, this method is guaranteed to *not* retain
        references to this main window on returning. References to child
        widgets (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        set_state_from_phase : CallableTypes
            Slot of the parent proactor setting the current state of the
            proactor to the current state of the passed simulator phase. For
            simplicity, this method iteratively for all simulator phases:

            * Connects the queueing signal emitted by this phase to this slot.
            * Calls this slot with this phase, ensuring that the proactor
              derive its initial state from the initial state of each phase.
        '''

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Log this finalization.
        logs.log_debug('Sanitizing simulator proactor phaser state...')

        # Initialize all phases (in arbitrary order).
        self._phase_seed.init(kind=SimPhaseKind.SEED, main_window=main_window)
        self._phase_init.init(kind=SimPhaseKind.INIT, main_window=main_window)
        self._phase_sim .init(kind=SimPhaseKind.SIM , main_window=main_window)

        # For each possible phase...
        for phase in self.PHASES:
            # Connect the queueing signal emitted by this phase to the
            # corresponding slot of the parent proactor.
            phase.queue_changed.connect(set_state_from_phase)

            # Set the proactor's initial state from this phase's initial state.
            set_state_from_phase(phase)

    # ..................{ GETTERS                           }..................
    def get_metadata(self) -> SimmerProactorMetadata:
        '''
        Named tuple aggregating metadata synopsizing the current state of this
        proactor, typically for displaying this metadata to the end user.

        Design
        ----------
        This method is intentionally designed as a getter rather than read-only
        property to inform callers of the non-negligible cost of each call to
        this getter, whose return value should thus be stored in a
        caller-specific variable rather than recreated silently on each access
        of such a property.
        '''

        # Number of simulator phases queued for modelling and exporting.
        phases_queued_modelling_count = 0
        phases_queued_exporting_count = 0

        # For each such phase...
        for phase in self.PHASES:
            # If this phase is queued for modelling, note this fact.
            if phase.is_queued_modelling:
                phases_queued_modelling_count += 1

            # If this phase is queued for exporting, note this fact.
            if phase.is_queued_exporting:
                phases_queued_exporting_count += 1

        # Create and return a named tuple aggregating this metadata.
        return SimmerProactorMetadata(
            phases_queued_modelling_count=phases_queued_modelling_count,
            phases_queued_exporting_count=phases_queued_exporting_count,
        )

    # ..................{ ENQUEUERS                         }..................
    def enqueue_phase_workers(self) -> QueueType:
        '''
        Create and return a new **simulator worker queue** (i.e., double-ended
        queue of each simulator worker to be subsequently run in a
        multithreaded manner by the parent proactor to run a simulation
        subcommand whose corresponding checkbox was checked at the time this
        method was called).

        This method enqueues (i.e., pushes onto this queue) workers in
        simulation phase order, defined as the ordering of the members of the
        :class:`betse.science.phase.phaseenum.SimPhaseKind` enumeration.
        Callers may safely run the simulation phases performed by these workers
        merely by sequentially assigning each worker enqueued in this queue to
        a thread via the
        :func:`betsee.util.thread.pool.guipoolthread.start_worker` function.

        For example:

        #. The :class:`QBetseeSimmerSubcommandWorkerModelSeed` worker (if any)
           is guaranteed to be queued *before*...
        #. The :class:`QBetseeSimmerSubcommandWorkerModelInit` worker (if any)
           is guaranteed to be queued *before*...
        #. The :class:`QBetseeSimmerSubcommandWorkerModelSim` worker (if any).

        Caveats
        ----------
        **This queue is double- rather than single-ended.** Why? Because the
        Python stdlib fails to provide the latter. Since the former generalizes
        the latter, however, leveraging the former in a single-ended manner
        replicates the behaviour of the latter. Ergo, a double-ended queue
        remains the most space- and time-efficient data structure for doing so.

        Returns
        ----------
        QueueType
            If either:

            * No simulator phase is currently queued.
            * Some simulator phase is currently running (i.e.,
              :attr:`_workers_queued` is already defined to be non-``None``).
        '''

        # Simulator worker queue to be classified *AFTER* creating and
        # enqueuing all workers.
        workers_queued = deque()

        # For each simulator phase...
        for phase in self.PHASES:
            # If this phase is currently queued for modelling...
            if phase.is_queued_modelling:
                # Simulator worker modelling this phase.
                worker = QBetseeSimmerPhaseWorker(
                    phase=phase, phase_subkind=SimmerPhaseSubkind.MODELLING)

                # Enqueue a new instance of this subclass.
                workers_queued.append(worker)

            # If this phase is currently queued for exporting...
            if phase.is_queued_exporting:
                # Simulator worker subclass exporting this phase.
                worker = QBetseeSimmerPhaseWorker(
                    phase=phase, phase_subkind=SimmerPhaseSubkind.EXPORTING)

                # Enqueue a new instance of this subclass.
                workers_queued.append(worker)

        # Return this queue.
        return workers_queued



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **pooled worker signals** (i.e., collection of all :class:`Signal`
instances thread-safely emittable by the :meth:`QBetseeThreadPoolWorker.run`
method from an arbitrary pooled thread possibly running *no* Qt event loop)
classes.
'''

#FIXME: To avoid thread and widget (de)synchronization issues, we should also
#connect the "paused" and "resumed" signals defined below to corresponding
#slots of the "QBetseeSimmer" controller.

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QObject, Signal
# from betse.util.io.log import logs
from betse.util.type.types import type_check, WeakRefBoundMethodType

# ....................{ SUPERCLASSES                      }....................
class QBetseeThreadPoolWorkerSignals(QObject):
    '''
    Low-level **pooled worker signals** (i.e., collection of all
    :class:`Signal` instances thread-safely emittable by the
    :meth:`QBetseeThreadPoolWorker.run` method from an arbitrary pooled thread
    possibly running *no* Qt event loop).

    Each instance of this class is owned by a pooled worker (i.e.,
    :class:`QBetseeThreadPoolWorker` instance), whose :meth:`run` method emits
    signals defined by this class typically connected to slots defined by
    objects residing in the original thread in which this worker was
    instantiated (e.g., the main event thread).

    Signal Emission
    ----------
    External callers attempting to emit signals defined by this collection
    should typically prefer to call the higher-level wrapper methods whose
    names are prefixed by ``emit_`` (e.g., :meth:`emit_progress_range`) in
    lieu of the associate lower-level signals (e.g., :attr:`progress_ranged`).
    The former wrap the latter with essential multithreading handling,
    including guaranteeably calling the vital :meth:`halt_work_if_requested`
    method *after* emitting each such signal. Nonetheless, to permit these
    signals to be trivially connected to, these signals necessarily remain
    public rather than private variables.

    Thread Affinity
    ----------
    Each instance of this class resides in the original thread in which this
    worker was instantiated and resides. Hence, neither this class nor any
    subclass of this class should define slots. Why? Qt would execute these
    slots in that original thread rather than the thread running this worker.

    Attributes
    ----------
    _halt_work_if_requested : WeakRefBoundMethodType
        Weak reference to a bound method either temporarily or permanently
        halting all business logic performed by the parent worker that owns
        this collection when requested to do so by external callers residing in
        other threads.
    '''

    # ..................{ SIGNALS                           }..................
    started = Signal()
    '''
    Signal emitted by the :meth:`QBetseeThreadPoolWorker.run` method
    immediately before running the :meth:`QBetseeThreadPoolWorker._work` method
    performing all subclass-specific business logic, passed the minimum.
    '''

    # ..................{ SIGNALS ~ progress                }..................
    progress_ranged = Signal(int, int)
    '''
    Signal optionally emitted by the subclass-specific
    :meth:`QBetseeThreadPoolWorker._work` method, passed the pair of integers
    ``progress_min`` and ``progress_max`` signifying the minimum and maximum
    **progress values** (i.e., integers subsequently emitted by the
    :attr:`progressed` signal) for the parent worker.

    Caveats
    ----------
    **This signal is neither guaranteed to be emitted nor guaranteed to be
    emitted in a timely manner.** This signal is only emitted at the optional
    discretion of subclass-specific business logic. Given this uncertainty,
    slots connected to this signal are advised to perform only non-essential
    logic (e.g., by calling only a :meth:`QProgressBar.setRange` method).

    This signal should be emitted by calling the higher-level
    :meth:`emit_progress_range` wrapper method rather than the lower-level
    :meth:`QSignal.emit` method of this signal.
    '''


    progress_stated = Signal(str)
    '''
    Signal optionally emitted by the subclass-specific
    :meth:`QBetseeThreadPoolWorker._work` method, passed the string
    ``status`` signifying the current state of progress as an arbitrary
    human-readable message for the parent worker.

    Caveats
    ----------
    **This signal is neither guaranteed to be emitted nor guaranteed to be
    emitted in a timely manner.** This signal is only emitted at the optional
    discretion of subclass-specific business logic. Given this uncertainty,
    slots connected to this signal are advised to perform only non-essential
    logic (e.g., by calling only a :meth:`QTextWidget.setText` method).

    This signal should be emitted by calling the higher-level
    :meth:`emit_progress_state` wrapper method rather than the lower-level
    :meth:`QSignal.emit` method of this signal.
    '''


    progressed = Signal(int)
    '''
    Signal optionally and repeatedly emitted by the subclass-specific
    :meth:`QBetseeThreadPoolWorker._work` method, passed an integer signifying
    the current progress of work completed by the parent worker.

    This integer is assumed to be in the range
    ``[progress_min, progress_max]``, where ``progress_min`` and
    ``progress_max`` are the pair of integers previously emitted from the
    :attr:`progress_ranged` signal (assuming the
    :meth:`QBetseeThreadPoolWorker._work` method previously did so).

    Caveats
    ----------
    **This signal is neither guaranteed to be emitted nor guaranteed to be
    emitted in a timely manner.** This signal is only emitted at the optional
    discretion of subclass-specific business logic. Thanks to this uncertainty,
    slots connected to this signal are advised to perform only non-essential
    logic (e.g., by calling only a :meth:`QProgressBar.setValue` method).

    This signal should be emitted by calling the higher-level
    :meth:`emit_progress` wrapper method rather than the lower-level
    :meth:`QSignal.emit` method of this signal.
    '''

    # ..................{ SIGNALS ~ paused                  }..................
    paused = Signal()
    '''
    Signal emitted by the :meth:`QBetseeThreadPoolWorker._block_work` method
    immediately before indefinitely blocking the parent worker.
    '''


    resumed = Signal()
    '''
    Signal emitted by the :meth:`QBetseeThreadPoolWorker._block_work` method
    immediately after unblocking the parent worker.
    '''

    # ..................{ SIGNALS ~ finished                }..................
    finished = Signal(bool)
    '''
    Signal emitted by the :meth:`QBetseeThreadPoolWorker.run` method
    immediately before returning from that method, passed either ``True`` if
    that method successfully performed all worker-specific business logic
    (i.e., if the :meth:`_work` method successfully returned *without* raising
    exceptions) *or* ``False`` otherwise.

    For finer-grained control over worker results, consider connecting instead
    to the:

    * :attr:`failed` signal for exceptions raised by the parent worker.
    * :attr:`succeeded` signal for objects returned by the parent worker.
    '''


    failed = Signal(Exception)
    '''
    Signal emitted by the :meth:`QBetseeThreadPoolWorker.run` method on
    catching a fatal exception raised by the subclass-specific
    :meth:`QBetseeThreadPoolWorker._work` method, passed this exception as is.

    Usage
    ----------
    Since this is Python 3.x rather than Python 2.x, slots connected to this
    signal may trivially reraise this exception from any other thread (complete
    with a contextual traceback specific to the stack for this thread) via the
    usual ``from`` clause of a ``raise`` statement.

    For this reason, this signal was intentionally designed *not* to emit the
    3-tuple returned by the standard :func:`sys.exc_info` function -- as would
    be required under Python 2.x to properly reraise this exception. Note that
    a little-known alternative to the ``from`` clause of a ``raise`` statement
    does technically exist: the :meth:`Exception.with_traceback` method. By
    explicitly calling this method on a newly instantiated exception passed
    `sys.exc_info()[2]`` (e.g., as ``raise
    MyNewException('wat?').with_traceback(sys.exc_info()[2])``), a similar
    effect is achievable. Since this is substantially less trivial, however,
    the prior approach is currently preferred.
    '''


    succeeded = Signal(object)
    '''
    Signal emitted by the :meth:`QBetseeThreadPoolWorker.run` method on
    successfully completing the parent worker, passed the arbitrary value
    returned by the subclass-specific :meth:`QBetseeThreadPoolWorker._work`
    method if that method returned a value *or* ``None`` otherwise (i.e., in
    the case that method returned no value).
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, halt_work_if_requested: WeakRefBoundMethodType) -> None:
        '''
        Initialize this pooled worker signals collection.

        Parameters
        ----------
        halt_work_if_requested : WeakRefBoundMethodType
            Weak reference to a bound method either temporarily or permanently
            halting all business logic performed by the parent worker that owns
            this collection when requested to do so by external callers
            residing in other threads.  For convenience, this is typically the
            :meth:`guipoolwork.QBetseeThreadPoolWorker._halt_work_if_requested`
            method bound to this parent worker. Accepting a reference to this
            method rather than this parent worker avoids circular object
            references between this parent worker and this child collection.
        '''

        # Initialize our superclass.
        super().__init__()

        # Classify all passed parameters.
        self._halt_work_if_requested = halt_work_if_requested

    # ..................{ EMITTERS ~ progress               }..................
    @type_check
    def emit_progress_range(
        self, progress_min: int, progress_max: int) -> None:
        '''
        Emit the :attr:`progress_ranged` signal with the passed range of all
        possible **progress values** (i.e., integers subsequently emitted by
        the :attr:`emit_progress` signal method) for the parent worker.

        Parameters
        ----------
        progress_min : int
            Minimum progress value emitted by :attr:`emit_progress`.
        progress_max : int
            Maximum progress value emitted by :attr:`emit_progress`.

        See Also
        ----------
        :attr:`progress_ranged`
            Further details.
        '''

        # Signal all slots connected to this signal with these parameters.
        self.progress_ranged.emit(progress_min, progress_max)

        # Temporarily or permanently halt all worker-specific business logic
        # when requested to do so by external callers in other threads *AFTER*
        # successfully emitting this signal.
        self._halt_work_if_requested()()


    @type_check
    def emit_progress_state(self, status: str) -> None:
        '''
        Emit the :attr:`progress_stated` signal with the passed **progress
        status** (i.e., string subsequently emitted by the
        :attr:`emit_progress_state` signal method) for the parent worker.

        Parameters
        ----------
        status : str
            Human-readable string signifying the progress of work completed.

        See Also
        ----------
        :attr:`progress_stated`
            Further details.
        '''

        # Signal all slots connected to this signal with these parameters.
        self.progress_stated.emit(status)

        # Temporarily or permanently halt all worker-specific business logic
        # when requested to do so by external callers in other threads *AFTER*
        # successfully emitting this signal.
        self._halt_work_if_requested()()


    @type_check
    def emit_progress(self, progress: int) -> None:
        '''
        Emit the :attr:`progressed` signal with the passed **progress value**
        (i.e., integer signifying the progress of work completed) for the
        parent worker.

        Parameters
        ----------
        progress : int
            Integer signifying the progress of work completed.

        See Also
        ----------
        :attr:`progressed`
            Further details.
        '''

        # Signal all slots connected to this signal with these parameters.
        self.progressed.emit(progress)

        # Temporarily or permanently halt all worker-specific business logic
        # when requested to do so by external callers in other threads *AFTER*
        # successfully emitting this signal.
        self._halt_work_if_requested()()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:class:`QCheckBox`-based simulation configuration widget subclasses.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, Signal
from PySide2.QtWidgets import QCheckBox
from betse.util.type.types import type_check, ClassOrNoneTypes
from betsee.gui.simconf.stack.widget.abc.guisimconfwdgeditscalar import (
    QBetseeSimConfEditScalarWidgetMixin)

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfCheckBox(QBetseeSimConfEditScalarWidgetMixin, QCheckBox):
    '''
    Simulation configuration-specific check box widget, permitting booleans
    backed by external simulation configuration files to be interactively
    edited.
    '''

    # ..................{ MIXIN ~ property : read-only       }..................
    @property
    def undo_synopsis(self) -> str:
        return QCoreApplication.translate(
            'QBetseeSimConfCheckBox', 'edits to a check box')


    @property
    def _finalize_widget_change_signal(self) -> Signal:
        return self.toggled


    @property
    def _sim_conf_alias_type_strict(self) -> ClassOrNoneTypes:
        return bool

    # ..................{ MIXIN ~ property : value           }..................
    @property
    def widget_value(self) -> bool:
        return self.isChecked()


    @widget_value.setter
    @type_check
    def widget_value(self, widget_value: bool) -> None:

        # Set this widget's displayed value to the passed value by calling the
        # setChecked() method of our superclass rather than this subclass,
        # preventing infinite recursion. (See the superclass method docstring.)
        super().setChecked(widget_value)


    def _reset_widget_value(self) -> None:
        self.widget_value = False



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based tree widget exposing all high-level features of a
simulation configuration.
'''

# ....................{ IMPORTS                           }....................
# from PySide2.QtCore import QCoreApplication  #, Slot
from PySide2.QtWidgets import QMainWindow
from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.util.widget.stock.guitreewdg import QBetseeTreeWidget

# ....................{ CLASSES                           }....................
class QBetseeSimConfTreeWidget(QBetseeTreeWidget):
    '''
    :mod:`PySide2`-based tree widget exposing all high-level features of the
    current simulation configuration.

    This application-specific widget augments the stock :class:`QTreeWidget`
    with support for handling simulation configurations, including:

    * Auto-axpansion of all tree items by default.
    * Integration with the corresponding :class:`QStackedWidget`, exposing all
      low-level configuration settings for the high-level simulation feature
      currently selected from this tree.
    '''

    # ..................{ INITIALIZERS                      }..................
    # def __init__(self, *args, **kwargs) -> None:
    #
    #     # Initialize our superclass with all passed parameters.
    #     super().__init__(*args, **kwargs)

    # To avoid circular import dependencies, this parameter is validated to be
    # an instance of the "QMainWindow" superclass rather than the expected
    # "QBetseeMainWindow" subclass of the "betsee.gui.window.guimainwindow"
    # submodule. Why? Because the latter imports the cached "betsee_ui.py"
    # module which imports the current submodule. Since this application only
    # contains one main window, this current validation suffices.
    @type_check
    def init(self, main_window: QMainWindow) -> None:
        '''
        Initialize this tree widget against the passed parent main window.

        This method is principally intended to perform **post-population
        initialization** (i.e., initialization performed *after* this widget
        has been completely pre-populated with all initial tree items).

        To avoid circular references, this method is guaranteed to *not* retain
        a reference to this main window on returning. References to child
        widgets (e.g., simulation configuration stack widget) of this window
        may be retained, however.

        Parameters
        ----------
        main_window: QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this widget.
        '''

        # Initialize our superclass.
        super().init()

        # Log this initialization.
        logs.log_debug('Initializing top-level tree widget...')

        # Initialize all widgets pertaining to the state of this simulation
        # configuration *BEFORE* connecting all relevant signals and slots
        # typically expecting these widgets to be initialized.
        self._init_widgets(main_window)
        self._init_connections(main_window)


    @type_check
    def _init_widgets(self, main_window: QMainWindow) -> None:
        '''
        Create all widgets owned directly by this tree widget *and* initialize
        all other widgets (not necessarily owned by this tree widget) whose
        internal state pertains to the high-level state of this tree widget.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this widget.
        '''

        # Sequence of all placeholder top-level placeholder items (i.e., items
        # whose corresponding stacked page has yet to be implemented) removed.
        # While extraneous, these items reside in the corresponding "betsee.ui"
        # file as a visual aid to streamline this transitional design phase.
        items_top_todo = []

        # For each top-level item of this tree widget...
        for item_top in self.iter_items_top():
            # If either:
            #
            # * One or more top-level placeholder items have been previously
            #   visited by this iteration, *ALL* top-level items from this item
            #   onward may be safely assumed to also be placeholder items.
            # * No top-level placeholder items have been visited yet *AND* this
            #   item's text implies this item to be the first such placeholder.
            #
            # Then this is a top-level placeholder item. In either case, append
            # this item to this sequence.
            if items_top_todo or item_top.text(0) == '--[TODO]--':
                items_top_todo.append(item_top)

        # Remove these items *AFTER* finding these items. While removing these
        # items during the above iteration would be preferable, doing so would
        # invite iteration desynchronization issues. Be safe... not sorry.
        for item_top_todo in items_top_todo:
            # Log this removal.
            logs.log_debug(
                'Removing top-level placeholder tree widget item "%s"...',
                item_top_todo.text(0))

            # Remove this item. Ideally, we would call the
            # guitreewdg.remove_item() function here. Sadly, that function
            # requires "shiboken2" functionality unavailable under non-standard
            # (but common) PySide2 installations.
            self.takeTopLevelItem(self.indexOfTopLevelItem(item_top_todo))

        # Expand all items of this tree widget to arbitrary depth *AFTER*
        # removing extraneous items above.
        self.expandAll()


    def _init_connections(self, main_window: QMainWindow) -> None:
        '''
        Connect all relevant signals and slots of this tree widget and the
        corresponding simulation configuration stack widget.

        Parameters
        ----------
        main_window: QBetseeMainWindow
            Parent :class:`QMainWindow` widget to initialize this widget with.
        '''

        # When an item of this tree widget is clicked, switch to the associated
        # page of this simulation configuration stack widget (if any).
        self.currentItemChanged.connect(
            main_window.sim_conf_stack.switch_page_to_tree_item)

        # First item of this tree widget.
        tree_item_first = self.topLevelItem(0)

        # Select this item *AFTER* connecting all relevant signals and slots,
        # ensuring that the corresponding slot is called.
        self.setCurrentItem(tree_item_first)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based directory functionality.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication
from PySide2.QtWidgets import QFileDialog
# from betse.util.io.log import logs
from betse.util.type.types import type_check, StrOrNoneTypes

# ....................{ SELECTORS                          }....................
#FIXME: Generalize to prevent users from traversing up (but *NOT* down) the
#directory tree during subdirectory selection. Doing so is non-trivial but
#absolutely feasible, requiring:
#
#* Definition of a new "QBetseeSubFileDialog" subclass of the stock
#  "QFileDialog" class.
#* In this subclass:
#  * Define a new "self._cwd_dirname" instance variable.
#  * Define a new enter_dir() @Slot.
#  * Connect this enter_dir() @Slot to the standard "directoryEntered" signal.
#  * In this @Slot (in order):
#    1. Test whether the passed dirname is prefixed by the desired dirname.
#    2. If so, set "self._cwd_dirname" to the passed dirname.
#    3. If not:
#       1. Preserve the current "self._cwd_dirname" as is.
#       2. Call "self.setDirectory(self._cwd_dirname)".
#* Manual instantiation and configuration of a "QBetseeSubFileDialog" instance
#  rather than simply calling the QFileDialog.getExistingDirectory() getter.
#
#It's actually pretty trivial. Make it so, please. Once we do, supply the
#resulting solution to the following StackOverflow answer -- which has yet to
#receive a single solution (but actually received the prototypical snarky "Why
#do you want to do this, because you never should."-style comment):
#    https://stackoverflow.com/questions/42001425/how-to-force-qfiledialog-getopenfilename-to-read-file-only-in-a-specific-folder
#    https://stackoverflow.com/questions/12169878/limit-directory-traversal-in-qfiledialog
#
#Full credit in our answer should go to the following quasi-answer:
#    https://stackoverflow.com/a/12173504/2809027

@type_check
def select_subdir(
    init_pathname: str, parent_dirname: str, *args, **kwargs) -> StrOrNoneTypes:
    '''
    Display a dialog requesting the user to select an existing subdirectory of
    the parent directory with the passed path, returning the relative pathname
    of this subdirectory with respect to this parent directory if this dialog
    was not cancelled *or* ``None`` otherwise (i.e., if this dialog was
    cancelled).

    Parameters
    ----------
    init_pathname : str
        Absolute or relative pathname of the subdirectory to initially display
        in this dialog. If this pathname is relative, this pathname is
        interpreted as relative to the ``parent_dirname`` parameter.
    parent_dirname : str
        Absolute pathname of the parent directory to select a subdirectory of.

    All remaining paremeters are passed as is to the :func:`guipath.select_path`
    function.

    Returns
    ----------
    StrOrNoneTypes
        Either:
        * If this dialog was confirmed, the absolute pathname of this
          subdirectory.
        * If this dialog was cancelled, ``None``.
    '''

    # Avoid circular import dependencies.
    from betsee.util.path import guipath

    # If no title was passed, default to a sensible title.
    if 'dialog_title' not in kwargs:
        kwargs['dialog_title'] = QCoreApplication.translate(
            'select_subdir', 'Select Subdirectory')

    # Return the user-based result of displaying this path dialog.
    return guipath.select_path(
        *args,
        dialog_callable=QFileDialog.getExistingDirectory,

        # Options with which to initialize this dialog. Specifically:
        #
        # * "ShowDirsOnly", both preventing users from selecting non-directory
        #   files and permitting Qt to display a native rather than non-native
        #   directory dialog. By default, this static function permits users to
        #   select any arbitrary file or directory *AND* prevents Qt from
        #   displaying a native dialog -- which, frankly, defeats the purpose.
        dialog_options=QFileDialog.ShowDirsOnly,

        # Passed pathnames unmodified.
        init_pathname=init_pathname,
        parent_dirname=parent_dirname,

        # Require these subdirectories to reside in this parent directory.
        is_subpaths=True,
        **kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Collection of the absolute paths of numerous critical files and directories
(both system-wide and user-specific) describing the structure of the local
filesystem in a general-purpose, cross-platform manner independent of this
application's specific usage of this filesystem.

See Also
----------
:mod:`betsee.guipathtree`
    Collection of the absolute paths of numerous critical files and directories
    describing the structure of this application on the local filesystem.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QStandardPaths
from betse.util.io.log import logs
from betse.util.os.shell import shelldir
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import type_check

# ....................{ GETTERS ~ dir                      }....................
#FIXME: Actually implement this function as described. For simplicity, this
#function currently defers to *ALWAYS* return get_user_docs_dirname(). Instead:
#
#* Define a new set_path_dialog_init_pathname() setter in this submodule,
#  internally caching the passed pathname to the application's "QSettings" store.
#* In the get_path_dialog_init_pathname() function:
#  * If the application's "QSettings" store contains this previously cached
#    pathname *AND* this pathname still exists, return this pathname.
#  * Else, return get_user_docs_pathname().
#* In all select_*() functions defined by the "guidir" and "guifile" submodules:
#  * *BEFORE* the path dialog is displayed and the caller passed no
#    "init_pathname" parameter (i.e., if this parameter is "None"), externally
#    call this getter to obtain the default value of this "init_pathname"
#    parameter. <---- O.K., this is now done.
#  * *AFTER* the end user successfully confirms this path dialog, externally
#    call this setter.
#
#Not terribly arduous and quite useful. Make it so, please.

def get_path_dialog_init_pathname() -> str:
    '''
    Absolute pathname of the path of arbitrary type (e.g., file, directory) to
    initially display in *path dialogs* (i.e., dialogs requesting the end user
    interactively select a possibly non-existing path).

    Returns
    ----------
    str
        Absolute pathname of either:
        * If the current user has already successfully selected at least one
          path from a path dialog _and_ the most recently selected such path
          still exists, that path.
        * Else (i.e., if this user has yet to select a path from a path dialog),
          a directory containing work-oriented files for this user.
    '''

    # Return the current user's documents directory.
    return get_user_docs_dirname()

# ....................{ GETTERS ~ dir : cached             }....................
@func_cached
def get_user_docs_dirname() -> str:
    '''
    Absolute pathname of the platform- and typically user-specific directory
    containing work-oriented files for the current user.

    This directory is:

    * On both Linux and macOS, ``~/Documents``.
    * On Windows, ``C:/Users/{USERNAME}/Documents``.
    '''

    return _get_dir(QStandardPaths.DocumentsLocation)

# ....................{ PRIVATE                            }....................
@type_check
def _get_dir(location: QStandardPaths.StandardLocation) -> str:
    '''
    Absolute path of the **standard directory** (i.e., platform- and typically
    user-specific directory) identified by the passed Qt enumeration constant.

    Parameters
    ----------
    location : QStandardPaths.StandardLocation
        Qt-specific enumeration constant identifying the standard directory to
        return this path of.

    Returns
    ----------
    str
        Absolute path of this standard directory.
    '''

    # Qt provides two means of querying for standard paths:
    #
    # * QStandardPaths.standardLocations(), a static getter function that may
    #   unsafely return the empty list "...if no locations for [the passed] type
    #   are defined."
    # * QStandardPaths.writableLocation(), a static getter function that may
    #   unsafely return "...an empty string if the location cannot be
    #   determined."
    #
    # Although both functions may return unsafe empty values, the latter is more
    # likely to do so and is thus less safe. Why? Because the former is
    # guaranteed to return at least as much and typically more than the latter.

    # List of the absolute paths of all directories satisfying this location,
    # sorted in descending order of presumed preference.
    dirnames = QStandardPaths.standardLocations(location)

    # If at least one such directory exists, defer to the absolute path of the
    # first such directory -- typically the same path returned by
    # "QStandardPaths.writableLocation(location)".
    if dirnames:
        dirname = dirnames[0]
    # Else, no such directories exist.
    else:
        # Log a non-fatal warning.
        logs.log_warning(
            'Standard location "%d" directories not found.', location)

        # For safety, fallback to the current working directory (CWD).
        dirname = shelldir.get_cwd_dirname()

    # Return this path.
    return dirname



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based stack widget page controllers specific to temporal
settings.
'''

# ....................{ IMPORTS                            }....................
#from PySide2.QtCore import QCoreApplication #, Signal, Slot
from PySide2.QtWidgets import QMainWindow
from betse.science.parameters import Parameters
# from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfTimeStackedWidgetPager(QBetseeControllerABC):
    '''
    :mod:`PySide2`-based stack widget page controller, connecting all editable
    widgets of the temporal page with the corresponding low-level settings of
    the current simulation configuration.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def init(self, main_window: QMainWindow) -> None:

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Simulation configuration state object.
        sim_conf = main_window.sim_conf

        # Initialize all initialization spin box widgets on this page.
        main_window.sim_conf_time_init_total.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.init_time_total)
        main_window.sim_conf_time_init_step.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.init_time_step)
        main_window.sim_conf_time_init_sampling.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.init_time_sampling)

        # Initialize all simulation spin box widgets on this page.
        main_window.sim_conf_time_sim_total.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.sim_time_total)
        main_window.sim_conf_time_sim_step.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.sim_time_step)
        main_window.sim_conf_time_sim_sampling.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.sim_time_sampling)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based stack widget page controllers specific to ions.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication #, Signal, Slot
from PySide2.QtWidgets import QMainWindow
from betse.science.parameters import Parameters
from betse.science.config.confenum import IonProfileType
# from betse.util.io.log import logs
from betse.util.type.types import type_check
from betse.util.type.mapping.mapcls import OrderedArgsDict
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfIonStackedWidgetPager(QBetseeControllerABC):
    '''
    :mod:`PySide2`-based stack widget page controller, connecting all editable
    widgets of the ion page with the corresponding low-level settings of the
    current simulation configuration.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def init(self, main_window: QMainWindow) -> None:

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Simulation configuration state object.
        sim_conf = main_window.sim_conf

        # Initialize all widgets on this page.
        main_window.sim_conf_ion_profile.init(
            sim_conf=sim_conf,
            sim_conf_alias=Parameters.ion_profile,
            enum_member_to_item_text=OrderedArgsDict(
                IonProfileType.BASIC, QCoreApplication.translate(
                    'QBetseeSimConfIonStackedWidgetPage', 'Basic'),
                IonProfileType.BASIC_CA, QCoreApplication.translate(
                    'QBetseeSimConfIonStackedWidgetPage', 'Basic + Ca2+'),
                IonProfileType.MAMMAL, QCoreApplication.translate(
                    'QBetseeSimConfIonStackedWidgetPage', 'Mammal'),
                IonProfileType.AMPHIBIAN, QCoreApplication.translate(
                    'QBetseeSimConfIonStackedWidgetPage', 'Amphibian'),
                IonProfileType.CUSTOM, QCoreApplication.translate(
                    'QBetseeSimConfIonStackedWidgetPage', 'Custom'),
            ),
        )

        #FIXME: Excise this after custom ion profile settings stabilize in the
        #BETSE core. When doing so:
        #
        #* Define a new _ion_profile_current_index_changed() slot in this class.
        #* Connect the "main_window.sim_conf_ion_profile.currentIndexChanged"
        #  signal to this slot.
        #* In the body of this slot:
        #  1. Add a one-liner disabling or enabling the
        #     "main_window.sim_conf_ion_custom" group box based on whether the
        #     newly selected ion profile type is the custom type or not.

        # Disable all currently non-working widgets on this page.
        main_window.sim_conf_ion_custom.setVisible(False)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulator** (i.e., :mod:`PySide2`-based object both displaying
*and* controlling the execution of simulation phases) functionality.
'''

#FIXME: Refactor this to leverage Qt's existing "QStateMachine" framework. Had
#we known about the existence of this framework, we likely wouldn't have
#implemented the cumbersome "QBetseeSimmerStatefulABC" API but instead leveraged
#the existing quite elegant "QStateMachine" API. Sadly, we know which of these
#two roads was taken. Unfortunately, refactoring this subpackage from the former
#to latter will probably be highly non-trivial -- but ultimately essential.
#
#There exist many, many reasons to transition to the "QStateMachine" API,
#including:
#
#* Sanity. We really just wanted a finite state machine (FSM). Now, we can. In
#  theory, an FSM-based implementation should be significantly saner.
#* Properties. The "QStateMachine" API permits the properties of arbitrary
#  "QObject" instances (like, say, whether or not a given "QPushButton" is
#  enabled or disabled) to be trivially set on transitioning to and from a
#  state. In theory, leveraging this should simplify our life with respect to
#  updating widget states.
#* History. The "QStateMachine" API provides the concept of "history states,"
#  which would permit the current state of the simulator to be saved and
#  restored (e.g., across desktop sessions). Probably essential, at some point.
#* Animations. The "QStateMachine" API provides the concept of "transition
#  animations," permitting the properties of arbitrary "QObject" instances to be
#  animated between... seemlessly. Nice, but hardly essential.
#
#Fortunately, Qt's documentation for this API is phenomenal. See the article
#entitled "The State Machine Framework." It's a surprisingly stunning read.

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QCoreApplication, Signal  #QObject, Slot
# from betse.util.io.log import logs
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.types import type_check  #, StrOrNoneTypes
from betsee.guiexception import BetseeSimmerException
from betsee.gui.simtab.run.guisimrunstate import SimmerState
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ SUPERCLASSES                      }....................
class QBetseeSimmerStatefulABC(QBetseeControllerABC):
    '''
    Abstract base class of all **stateful simulator controller** (i.e.,
    :mod:`PySide2`-based object controlling the internal and possibly external
    state of some aspect of the simulator) subclasses.

    Attributes (Private: Non-widgets)
    ----------
    _state : SimmerState
        Current state of this simulator controller, exactly analogous to the
        current state of a finite state automata. For safety, this variable
        should *only* be set by the public :meth:`state` setter.
    '''

    # ..................{ SIGNALS                           }..................
    state_changed = Signal(object, object)
    '''
    Signal passed the current and previous simulator states on each change to
    the state of this simulator controller, either due to user interaction or
    programmatic non-interaction.

    Caveats (Value)
    ----------
    **Note that the two passed states may be the same state.** This slot is
    unconditionally signalled on each attempt to assign this controller a new
    state, including those implicitly preserving the current state by
    reassigning this controller the same state. Why? Because slots connected to
    this signal expect to be called on each state assignment even if that
    assignment preserves that existing state. Why? Because the mere act of
    assignment signifies that lower-level metadata synopsizing this high-level
    state *may* have changed, despite this state *not* having changed.

    The canonical example is the total number of queued simulator phases. When
    the user queues any simulator phase *or* unqueues any simulator phase
    excluding the last queued phase, the proactor remains in the same queued
    state while the number of queued simulator phases changes. If the proactor
    did *not* unconditionally signal this slot, then:

    * The proactor would need to define a new signal and emit that signal on
      each change to the number of queued simulator phases.
    * The simulator would need to define a new slot and connect that slot to
      that signal to update its display of this number.

    Instead, this slot is unconditionally signalled.

    Caveats (Type)
    ----------
    **This signal should only be passed states of type :class:`SimmerState`,**
    despite accepting objects of arbitrary type. Ideally, this signal would be
    refactored to unambiguously accept objects of type :class:`SimmerState`.
    Sadly, doing so induces non-human-readable segmentation faults resembling:

    .. code::

        Fatal Python error: Segmentation fault

        Current thread 0x00007f2d83afd540 (most recent call first):
        File "./betsee/gui/simtab/run/guisimrunabc.py", line 63 in QBetseeSimmerStatefulABC
        File "./betsee/gui/simtab/run/guisimrunabc.py", line 48 in <module>
        File "<frozen importlib._bootstrap>", line 222 in _call_with_frames_removed
        File "<frozen importlib._bootstrap_external>", line 697 in exec_module
        File "<frozen importlib._bootstrap>", line 673 in _load_unlocked
        File "<frozen importlib._bootstrap>", line 957 in _find_and_load_unlocked
        File "<frozen importlib._bootstrap>", line 968 in _find_and_load
        File "./betsee/gui/simtab/run/guisimrunact.py", line 36 in <module>
        File "<frozen importlib._bootstrap>", line 222 in _call_with_frames_removed
        File "<frozen importlib._bootstrap_external>", line 697 in exec_module
        File "<frozen importlib._bootstrap>", line 673 in _load_unlocked
        File "<frozen importlib._bootstrap>", line 957 in _find_and_load_unlocked
        File "<frozen importlib._bootstrap>", line 968 in _find_and_load
        File "./betsee/gui/simtab/run/guisimrun.py", line 49 in <module>
        File "<frozen importlib._bootstrap>", line 222 in _call_with_frames_removed
        File "<frozen importlib._bootstrap_external>", line 697 in exec_module
        File "<frozen importlib._bootstrap>", line 673 in _load_unlocked
        File "<frozen importlib._bootstrap>", line 957 in _find_and_load_unlocked
        File "<frozen importlib._bootstrap>", line 968 in _find_and_load
        File "./betsee/gui/simtab/guisimtab.py", line 48 in __init__
        File "<string>", line 5 in func_type_checked
        File "./betsee/data/py/betsee_ui.py", line 1988 in setupUi
        File "./betsee/gui/window/guimainwindow.py", line 184 in __init__
        File "<string>", line 25 in func_type_checked
        File "./betsee/gui/guimain.py", line 313 in _make_main_window
        File "./betsee/gui/guimain.py", line 276 in run
        File "./betsee/cli/guicli.py", line 240 in _do
        File "/home/leycec/py/betse/betse/util/py/pyprofile.py", line 130 in _profile_callable_none
        File "/home/leycec/py/betse/betse/util/py/pyprofile.py", line 114 in profile_callable
        File "<string>", line 65 in func_type_checked
        File "/home/leycec/py/betse/betse/util/cli/cliabc.py", line 180 in run
        File "<string>", line 15 in func_type_checked
        File "./betsee/__main__.py", line 98 in main
        File "/usr/bin/betsee", line 66 in <module>
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this stateful simulator controller.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Default this controller's state to the unqueued state.
        self._state = SimmerState.UNQUEUED

    # ..................{ PROPERTIES ~ state                }..................
    @property
    def state(self) -> SimmerState:
        '''
        Current state of this simulator controller, exactly analogous to the
        current state of a finite state automata.
        '''

        return self._state


    @state.setter
    @type_check
    def state(self, state: SimmerState) -> None:
        '''
        Set the current state of this simulator controller to the passed state
        *and* signal all slots connected to the :attr:`state_changed` of
        this state change.
        '''

        # New and old states to be passed to all slots connected to the
        # "state_changed" signal *BEFORE* re-assigning these states.
        state_new = state
        state_old = self._state

        # Set the current state of this simulator controller to this state.
        self._state = state_new

        # Update the current state of both this simulator controller and
        # widgets controlled by this controller given this state.
        self._update_state()

        # Signal all connected slots *AFTER* internally updating this state.
        self.state_changed.emit(state_new, state_old)

    # ..................{ EXCEPTIONS                        }..................
    def _die_unless_queued(self) -> None:
        '''
        Raise an exception unless this stateful simulator controller is
        currently queued for modelling and/or exporting one or more simulator
        phases.

        Equivalently, this method raises an exception if *no* such phase is
        currently queued.

        See Also
        ----------
        :meth:`is_queued`
            Further details.
        '''

        if not self.is_queued:
            raise BetseeSimmerException(QCoreApplication.translate(
                'QBetseeSimmerStatefulABC',
                'Simulator controller not queued.'))

    # ..................{ SUBCLASS ~ properties             }..................
    # Abstract read-only properties required to be overridden by subclasses.

    @abstractproperty
    def is_queued(self) -> bool:
        '''
        ``True`` only if this stateful simulator controller is currently queued
        for modelling and/or exporting one or more simulator phases.
        '''

        pass

    # ..................{ SUBCLASS ~ methods                }..................
    # Concrete methods intended (but *NOT* required) to be overridden by
    # subclasses.

    def _update_state(self) -> None:
        '''
        Update the internal state of this stateful simulator controller and the
        contents of widgets controlled by this controller to reflect the
        current external state of this controller.

        The :meth:`state` setter property method internally calls this method
        to perform subclass-specific business logic on either the user
        interactively *or* the codebase programmatically interacting with any
        widget relevant to the current state of this controller (e.g., a
        checkbox queueing a simulation phase for exporting).

        Design
        ----------
        The default implementation of this method reduces to a noop and is thus
        intended (but *not* required) to be overridden by subclasses requiring
        subclass-specific business logic to be performed when this controller's
        :meth:`state` property is set. Overriding setter property methods in a
        manner internally calling the superclass method is highly non-trivial;
        hence, the :meth:`state` setter property method internally calls this
        method instead.
        '''

        pass



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **thread enumeration** (e.g., :class:`enum.Enum` subclass describing
different types of multithreading behaviour) functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.enums import make_enum

# ....................{ ENUMERATIONS                      }....................
ThreadWorkerState = make_enum(
    class_name='ThreadWorkerState',
    member_names=('IDLE', 'RUNNING', 'PAUSED', 'DELETED',))
'''
Enumeration of all supported types of **multithreaded worker state** (i.e.,
mutually exclusive execution state of a :class:`QObject`- or
:class:`QRunnable`-derived worker object, analogous to a state in a finite
state automata).

Attributes
----------
IDLE : enum
    Idle state, implying this worker to be **idle** (i.e., neither working nor
    paused while working). From this state, this worker may freely transition
    to the working but *not* paused state.
RUNNING : enum
    Running state, implying this worker to be **running** (i.e., performing
    subclass-specific business logic, typically expected to be long-running).
    From this state, this worker may freely transition to *any* other state.
PAUSED : enum
    Paused state, implying this worker to be **paused** (i.e., temporarily
    halted from performing subclass-specific business logic). From this state,
    this worker may freely transition to *any* other state. Transitioning from
    this state to the running state is also referred to as "resuming."
DELETED : enum
    Deleted state, implying this worker to be scheduled for deletion by a
    recent call to a ``delete_later`` method defined on this worker (e.g., the
    :meth:`QBetseeThreadPoolWorker.delete_later` method). From this state, this
    worker cannot freely transition to *any* other state.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all simulation configuration widget subclasses
instantiated in pages of the top-level stack.
'''

# ....................{ IMPORTS                            }....................
from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ MIXINS                             }....................
# To avoid metaclass conflicts with the "QWidget" base class inherited by all
# widgets also inheriting this base class, this base class *CANNOT* be
# associated with another metaclass (e.g., "abc.ABCMeta").
class QBetseeWidgetMixinSimConf(QBetseeObjectMixin):
    '''
    Abstract base class of all **non-editable simulation configuration widget**
    (i.e., widget *not* interactively editing simulation configuration values
    stored in external YAML files) subclasses.

    Design
    ----------
    This class is suitable for use as a multiple-inheritance mixin. To preserve
    the expected method resolution order (MRO) semantics, this class should
    typically be subclassed *first* rather than *last* in subclasses.

    Attributes
    ----------
    _sim_conf : QBetseeSimConf
        High-level state of the currently open simulation configuration, which
        depends on the state of this low-level simulation configuration widget.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed arguments.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._sim_conf = None


    @type_check
    def init(
        self,
        # To avoid circularity from the "QBetseeSimConf" class importing the
        # "QBetseeMainWindowConfig" class importing the "betsee_ui" submodule
        # importing instances of this class, this type is validated dynamically.
        sim_conf: 'betsee.gui.simconf.guisimconf.QBetseeSimConf',
    ) -> None:
        '''
        Finalize the initialization of this widget.

        Design
        ----------
        Subclasses should typically override this method with a
        subclass-specific implementation that (in order):

        #. Calls this superclass implementation, which sets instance variables
           typically required by subclass slots.
        #. Connects all relevant signals and slots.

        Connecting these signals and slots earlier in the :meth:`__init__`
        method is *not* recommended, even for slots that Qt technically should
        *never* invoke at that time. Why? Because Qt actually appears to
        erroneously emit signals documented as emitted only by external user
        action (e.g., :meth:`QLineEdit.editingFinished`) on internal code-based
        action (e.g., startup construction of the main window).

        Parameters
        ----------
        sim_conf : QBetseeSimConf
            High-level state of the currently open simulation configuration.
        '''

        # Log this initialization *AFTER* storing this name.
        logs.log_debug(
            'Initializing non-editable widget "%s"...', self.obj_name)

        # Classify all passed parameters. Since the main window rather than this
        # state object owns this widget, retaining a reference to this state
        # object introduces no circularity and hence is safe. (That's nice.)
        self._sim_conf = sim_conf



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Application **status bar** (i.e., :class:`QStatusBar` widget synopsizing
application status in the :class:`QMainWindow` singleton for this application)
facilities.
'''

# ....................{ IMPORTS                            }....................
# from PySide2.QtCore import QCoreApplication
from PySide2.QtWidgets import QStatusBar
# from betse.util.io.log import logs
from betse.util.type.types import type_check
# from betsee.guiexception import BetseePySideWindowException

# ....................{ GETTERS                            }....................
def get_status_bar() -> QStatusBar:
    '''
    Singleton status bar widget for this application.

    Specifically, this function returns either:

    * If this main window defines an instance variable named ``status_bar``,
      the value of this variable.
    * Else, the value returned by the :meth:`QMainWindow.statusBar` method.
      Since this method is `considered problematic`_ by some in the Qt
      development community, this method is *only* called as a fallback.

    .. _considered problematic:
       https://plashless.wordpress.com/2013/09/14/qt-qmainwindow-statusbar-dont-use-it

    Design
    ----------
    To avoid circular import dependencies, this getter intentionally resides in
    this submodule known *not* to be subject to these dependencies rather than
    in an arguably more germain submodule known to be subject to these
    dependencies (e.g., :mod:`betsee.gui.window.guimainwindow`).

    Returns
    ----------
    QStatusBar
        This widget.

    Raises
    ----------
    BetseePySideWindowException
        If the main window widget for this application is uninstantiated.
    '''

    # Avoid circular import dependencies.
    from betsee.util.app import guiappwindow

    # Main window singleton.
    main_window = guiappwindow.get_main_window()

    # Application-specific status bar widget defined by this window if any *OR*
    # "None" otherwise.
    status_bar = getattr(main_window, 'status_bar', None)

    # If this window defines no such widget, fallback to the
    # application-agnostic status bar defined by *ALL* "QMainWindow" widgets
    if status_bar is None:
        status_bar = main_window.statusBar()

    # Return this status bar.
    return status_bar

# ....................{ STATUS                             }....................
@type_check
def show_status(text: str) -> None:
    '''
    Display the passed string as a **temporary message** (i.e., string
    temporarily replacing any normal message currently displayed) in the
    status bar.
    '''

    # Singleton status bar widget for this application.
    status_bar = get_status_bar()

    #FIXME: Validate this string to contain no newlines. Additionally, consider
    #emitting a warning if the length of this string exceeds a sensible maximum
    #(say, 160 characters or so).

    # Display this temporary message with no timeout.
    status_bar.showMessage(text)


def clear_status() -> None:
    '''
    Remove the temporary message currently displayed in the status bar if any
    *or* reduce to a noop otherwise.

    This function restores any "permanent" message displayed in the status bar
    (if any) prior to the recent temporary message displayed in the status bar
    (if any) by calls to the :func:`show_status` function. (It's complicated.)
    '''

    # Singleton status bar widget for this application.
    status_bar = get_status_bar()

    # Remove the temporary message currently displayed in the status bar if any.
    status_bar.clearMessage()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level :mod:`PySide2`-based mouse cursor facilities.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import Qt
from PySide2.QtGui import QApplication
from betse.util.type.types import GeneratorType
from contextlib import contextmanager

# ....................{ CONTEXTS                           }....................
@contextmanager
def waiting_cursor() -> GeneratorType:
    '''
    Context manager changing the mouse cursor to the prototypical wait cursor
    (e.g., animated hourglass) for the duration of this context.

    This context manager guaranteeably reverts the cursor to the prior cursor
    even when fatal exceptions are raised (e.g., by the caller's block).

    Returns
    -----------
    contextlib._GeneratorContextManager
        Context manager changing the cursor as described above.

    Yields
    -----------
    None
        Since this context manager yields no values, the caller's ``with``
        statement must be suffixed by *no* ``as`` clause.
    '''

    # Attempt to...
    try:
        # Change the cursor to the wait cursor.
        QApplication.setOverrideCursor(Qt.WaitCursor)

        # Yield control to the body of the caller's "with" block.
        yield
    # Revert to the prior cursor even if that block raised an exception.
    finally:
        QApplication.restoreOverrideCursor()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **simulator worker** (i.e., :mod:`PySide2`-based thread performing
the equivalent of a simulation subcommand in a Qt-aware manner) functionality.
'''

# ....................{ IMPORTS                           }....................
# from PySide2.QtCore import QCoreApplication  # Slot, Signal
from betse.science.parameters import Parameters
from betse.science.simrunner import SimRunner
# from betse.util.io.log import logs
from betse.util.type import enums
from betse.util.type.cls import classes
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.obj import objects
from betse.util.type.types import type_check, CallableTypes
from betsee.guiexception import BetseePySideThreadWorkerException
from betsee.gui.simtab.run.guisimrunenum import SimmerState
from betsee.gui.simtab.run.phase.guisimrunphase import QBetseeSimmerPhase
from betsee.gui.simtab.run.work.guisimrunworkenum import SimmerPhaseSubkind
from betsee.gui.simtab.run.work.guisimrunworksig import SimCallbacksSignaller
from betsee.util.thread.pool.guipoolwork import QBetseeThreadPoolWorker

# ....................{ SUPERCLASSES                      }....................
class QBetseeSimmerWorkerABC(QBetseeThreadPoolWorker):
    '''
    Abstract base class of all low-level **simulator worker** (i.e.,
    worker running one or more arbitrary simulation operations) subclasses.

    Attributes
    ----------
    _conf_filename : StrOrNoneTypes
        Absolute filename of the YAML-formatted simulation configuration file
        defining the simulation to be run by this worker if the :meth:`init`
        method of this worker has been called *or* ``None`` otherwise.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self) -> None:
        '''
        Initialize this simulator worker.
        '''

        # Initialize our superclass.
        super().__init__()

        # Nullify all instance variables for safety.
        self._conf_filename = None


    @type_check
    def init(self, conf_filename: str, *args, **kwargs) -> None:
        '''
        Finalize the initialization of this simulator worker.

        Parameters
        ----------
        conf_filename : str
            Absolute filename of the YAML-formatted simulation configuration
            file defining the simulation to be run by this worker.

        All remaining parameters are passed as is to the superclass
        :meth:`init` method.
        '''

        # Finalize our superclass initialization with all passed parameters.
        super().init(*args, **kwargs)

        # Classify all passed parameters.
        self._conf_filename = conf_filename

    # ..................{ EXCEPTIONS                        }..................
    def _die_unless_initted(self) -> None:
        '''
        Raise an exception unless the :meth:`init` method has been called.

        Raises
        ----------
        BetseePySideThreadWorkerException
            If the :attr:`_conf_filename` instance variable is ``None``,
            in which case the :meth:`init` method has yet to be called.
        '''

        # If the filename of this worker's simulation configuration file has
        # yet to be set, the init() method has yet to be called. In this case,
        # raise an exception.
        if self._conf_filename is None:
            raise BetseePySideThreadWorkerException(
                '"{}" uninitialized (i.e., init() method not called).'.format(
                    objects.get_class_name(self)))

    # ..................{ MAKERS                            }..................
    def _make_sim_runner(self) -> SimRunner:
        '''
        Create and return a new **simulation phase runner** (i.e., high-level
        object encapsulating the running of simulation phases as corresponding
        public methods commonly referred to as simulation subcommands), whose
        thread affinity is that of the caller.

        Caveats
        ----------
        To ensure that this runner is owned by the thread running the
        :meth:`run` and :meth:`_work` methods, this method should be called
        *only* from the :meth:`_work` method.

        This method is intentionally implemented to create and return a new
        runner whose thread affinity is that of the caller. If this method were
        instead implemented to classify this runner is an instance variable of
        this superclass (e.g., ``self._sim_runner``), the thread affinity of
        this runner would erroneously be that of the main event thread.
        '''

        # Raise an exception unless the :meth:`init` method has been called.
        self._die_unless_initted()

        # Simulation configuration whose thread affinity is that of the caller.
        p = self._make_p()

        # Simulation phase callbacks whose thread affinity is that of the
        # caller.
        callbacks = SimCallbacksSignaller(signals=self.signals)

        # Simulation phase runner whose thread affinity is that of the caller.
        sim_runner = SimRunner(p=p, callbacks=callbacks)

        # Return this runner.
        return sim_runner


    def _make_p(self) -> Parameters:
        '''
        Create and return a deep copy of the singleton simulation
        configuration (i.e., :attr:`QBetseeMainWindow.sim_conf.p`), whose
        thread affinity is that of the caller.

        This method bridges the critical gap between the low-level CLI-based
        simulation runner and this high-level GUI-based simulator, making this
        the most important callable in either codebase.

        Design
        ----------
        For thread-safety, this worker returns a deep copy of the singleton
        simulation configuration rather than a shallow reference to that
        simulation configuration; ergo, that object is guaranteed to remain
        unchanged by this worker.

        Doing so also permits this method to dynamically reconfigure this
        configuration in-memory to satisfy GUI requirements, including:

        * Disable all simulation configuration options either requiring
          interactive user input *or* displaying graphical output intended for
          interactive user consumption (e.g., plots, animations).
        * Enable all simulation configuration options exporting to disk.

        Returns
        ----------
        Parameters
            Deep copy of the singleton simulation configuration (i.e.,
            :attr:`QBetseeMainWindow.sim_conf.p`), whose thread affinity is
            that of the caller.
        '''

        # Simulation configuration newly deserialized from the passed original
        # Simulation configuration.
        p = Parameters.make(self._conf_filename)

        # Disable all simulation configuration options either requiring
        # interactive user input *OR* displaying graphical output intended for
        # interactive user consumption (e.g., plots, animations).
        p.anim.is_after_sim_show = False
        p.anim.is_while_sim_show = False
        p.plot.is_after_sim_show = False

        # Enable all simulation configuration options exporting to disk.
        p.anim.is_after_sim_save = True
        p.anim.is_while_sim_save = True
        p.plot.is_after_sim_save = True

        # Return this configuration.
        return p

# ....................{ SUBCLASSES                        }....................
class QBetseeSimmerPhaseWorker(QBetseeSimmerWorkerABC):
    '''
    Low-level **simulator phase worker** (i.e., simulator worker running an
    arbitrary simulation phase).

    Attributes
    ----------
    _phase : QBetseeSimmerPhase
        Simulator phase run by this worker.
    _phase_subkind : SimmerPhaseSubkind
        Type of work performed within this phase by this worker.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,
        phase: QBetseeSimmerPhase,
        phase_subkind: SimmerPhaseSubkind,
    ) -> None:
        '''
        Initialize this simulator phase worker.

        Parameters
        ----------
        phase : QBetseeSimmerPhase
            Simulator phase run by this worker.
        phase_subkind : SimmerPhaseSubkind
            Type of work performed within this phase by this worker.
        '''

        # Initialize our superclass.
        super().__init__()

        # Classify all passed parameters.
        self._phase = phase
        self._phase_subkind = phase_subkind

    # ..................{ PROPERTIES                        }..................
    # Read-only concrete properties.

    @property
    def phase(self) -> QBetseeSimmerPhase:
        '''
        Simulator phase run by this worker.
        '''

        return self._phase


    @property
    def phase_subkind(self) -> SimmerPhaseSubkind:
        '''
        Type of work performed within the simulator phase run by this worker.
        '''

        return self._phase_subkind


    @property_cached
    def simmer_state(self) -> SimmerState:
        '''
        Type of work performed within the simulator phase run by this worker as
        a member of the :class:`SimmerState` rather than
        :class:`SimmerPhaseSubkind` enumeration.
        '''

        return enums.get_member_from_value(
            enum_type=SimmerState,
            enum_member_value=self._phase_subkind.value)

    # ..................{ GETTERS                           }..................
    def _get_sim_runner_subcommand(self) -> CallableTypes:
        '''
        **Simulation subcommand** (i.e., public method of the
        :class:`SimRunner` class) run by this worker's :meth:`work` method.

        The :meth:`_work` method internally invokes this subcommand with the
        local :class:`SimRunner` instance created and returned by the
        :meth:`_make_sim_runner` method.

        Design
        ----------
        This callable is intentionally implemented as a getter rather than
        property. Although properties do largely obsolete getters, refactoring
        this getter into a property introduces awkward implicit calling
        semantics inviting spurious linter errors. Ergo, explicit is better
        than implicit in this edge case.
        '''

        # "_"-suffixed substring prefixing the name of the unbound "SimRunner"
        # method called by this worker if any *OR* the empty string otherwise.
        method_name_prefix = (
            'plot_' if self.phase_subkind is SimmerPhaseSubkind.EXPORTING else
            '')

        # Name of this method.
        method_name = method_name_prefix + self.phase.name

        # Return this method.
        return classes.get_method(cls=SimRunner, method_name=method_name)

    # ..................{ WORKERS                           }..................
    def _work(self) -> None:

        # Simulation phase runner whose thread affinity is that of the caller.
        sim_runner = self._make_sim_runner()

        # Simulation subcommand to be run by this worker.
        sim_runner_subcommand = self._get_sim_runner_subcommand()

        # Run this subcommand on this runner.
        sim_runner_subcommand(sim_runner)
